@@ -1,3 +1,8 @@
+/*
+ * Nail - a mail user agent derived from Berkeley Mail.
+ *
+ * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ */
 /*
  * Copyright (c) 1980, 1993
  *	The Regents of the University of California.  All rights reserved.
@@ -33,7 +38,7 @@
 
 #ifndef lint
 #ifdef	DOSCCS
-static char sccsid[] = "@(#)cmd3.c	1.9 (gritter) 5/22/02";
+static char sccsid[] = "@(#)cmd3.c	2.15 (gritter) 11/8/02";
 #endif
 #endif /* not lint */
 
@@ -45,7 +50,21 @@ static char sccsid[] = "@(#)cmd3.c	1.9 (gritter) 5/22/02";
  *
  * Still more user commands.
  */
-static int diction __P((const void *, const void *));
+static int	diction __P((const void *, const void *));
+static int	bangexp __P((char **, size_t *));
+static void	make_ref __P((struct message *, struct header *));
+static int	respond_internal __P((int *, int));
+static char	*reedit __P((char *));
+static void	sort __P((char **));
+static int	Respond_internal __P((int [], int));
+static int	forward1 __P((void *, int));
+static RETSIGTYPE	onpipe __P((int));
+static enum okay	delete_shortcut __P((const char *));
+#ifdef	__STDC__
+static int	(*respond_or_Respond(int c))(int *, int);
+#else
+static int	(*respond_or_Respond())();
+#endif
 
 /*
  * Process a shell escape by saving signals, ignoring signals,
@@ -56,19 +75,21 @@ shell(v)
 	void *v;
 {
 	char *str = v;
-	signal_handler_t sigint = safe_signal(SIGINT, SIG_IGN);
+	sighandler_type sigint = safe_signal(SIGINT, SIG_IGN);
 	char *shell;
-	char cmd[BUFSIZ];
+	char *cmd;
+	size_t cmdsize;
 
-	(void) strncpy(cmd, str, BUFSIZ);
-	cmd[BUFSIZ-1]='\0';
-	if (bangexp(cmd, BUFSIZ) < 0)
+	cmd = smalloc(cmdsize = strlen(str) + 1);
+	strcpy(cmd, str);
+	if (bangexp(&cmd, &cmdsize) < 0)
 		return 1;
 	if ((shell = value("SHELL")) == NULL)
 		shell = PATH_CSHELL;
 	(void) run_command(shell, 0, -1, -1, "-c", cmd, NULL);
 	(void) safe_signal(SIGINT, sigint);
 	printf("!\n");
+	free(cmd);
 	return 0;
 }
 
@@ -80,14 +101,14 @@ int
 dosh(v)
 	void *v;
 {
-	signal_handler_t sigint = safe_signal(SIGINT, SIG_IGN);
+	sighandler_type sigint = safe_signal(SIGINT, SIG_IGN);
 	char *shell;
 
 	if ((shell = value("SHELL")) == NULL)
 		shell = PATH_CSHELL;
 	(void) run_command(shell, 0, -1, -1, NULL, NULL, NULL);
 	(void) safe_signal(SIGINT, sigint);
-	putchar('\n');
+	sputc('\n', stdout);
 	return 0;
 }
 
@@ -96,59 +117,62 @@ dosh(v)
  * last issued command where possible.
  */
 
-char	lastbang[128];
+static char	*lastbang;
+static size_t	lastbangsize;
 
-int
+static int
 bangexp(str, size)
-	char *str;
-	int size;
+	char **str;
+	size_t *size;
 {
-	char bangbuf[BUFSIZ];
-	char *cp, *cp2;
-	int n;
+	char *bangbuf;
 	int changed = 0;
+	int dobang = value("bang") != NULL;
+	size_t sz, i, j, bangbufsize;
 
-	cp = str;
-	cp2 = bangbuf;
-	n = BUFSIZ;
-	while (*cp) {
-		if (*cp == '!') {
-			if (n < strlen(lastbang)) {
-overf:
-				printf("Command buffer overflow\n");
-				return(-1);
+	bangbuf = smalloc(bangbufsize = *size);
+	i = j = 0;
+	while ((*str)[i]) {
+		if (dobang) {
+			if ((*str)[i] == '!') {
+				sz = strlen(lastbang);
+				bangbuf = srealloc(bangbuf, bangbufsize += sz);
+				changed++;
+				strcpy(&bangbuf[j], lastbang);
+				j += sz;
+				i++;
+				continue;
 			}
-			changed++;
-			strcpy(cp2, lastbang);
-			cp2 += strlen(lastbang);
-			n -= strlen(lastbang);
-			cp++;
-			continue;
 		}
-		if (*cp == '\\' && cp[1] == '!') {
-			if (--n <= 1)
-				goto overf;
-			*cp2++ = '!';
-			cp += 2;
+		if ((*str)[i] == '\\' && (*str)[i + 1] == '!') {
+			bangbuf[j++] = '!';
+			i += 2;
 			changed++;
 		}
-		if (--n <= 1)
-			goto overf;
-		*cp2++ = *cp++;
+		bangbuf[j++] = (*str)[i++];
 	}
-	*cp2 = 0;
+	bangbuf[j] = '\0';
 	if (changed) {
 		printf("!%s\n", bangbuf);
 		fflush(stdout);
 	}
-	strncpy(str, bangbuf, size);
-	str[size-1]='\0';
-	strncpy(lastbang, bangbuf, 128);
-	lastbang[127] = 0;
+	sz = j;
+	if (sz >= *size)
+		*str = srealloc(*str, *size = sz + 1);
+	strcpy(*str, bangbuf);
+	if (sz >= lastbangsize)
+		lastbang = srealloc(lastbang, lastbangsize = sz + 1);
+	strcpy(lastbang, bangbuf);
+	free(bangbuf);
 	return(0);
 }
 
-static const char helptext[] =
+/*ARGSUSED*/
+int
+help(v)
+	void *v;
+{
+	const char *helptext = catgets(catd, CATSET, 36,
 "    Nail   Commands\n\
 t <message list>                type messages\n\
 n                               goto and type next message\n\
@@ -159,7 +183,7 @@ s <message list> file           append messages to file\n\
 u <message list>                undelete messages\n\
 R <message list>                reply to message senders\n\
 r <message list>                reply to message senders and all recipients\n\
-pre <message list>              make messages go back to /var/spool/mail\n\
+pre <message list>              make messages go back to /var/mail\n\
 m <user list>                   mail to specific users\n\
 q                               quit, saving unresolved messages in mbox\n\
 x                               quit, do not remove system mailbox\n\
@@ -171,13 +195,8 @@ A <message list> consists of integers, ranges of same, or user names separated\n
 by spaces.  If omitted, Nail uses the last message typed.\n\
 \n\
 A <user list> consists of user names or aliases separated by spaces.\n\
-Aliases are defined in .mailrc in your home directory.\n";
+Aliases are defined in .mailrc in your home directory.\n");
 
-/*ARGSUSED*/
-int
-help(v)
-	void *v;
-{
 	fputs(helptext, stdout);
 	return(0);
 }
@@ -204,7 +223,7 @@ schdir(v)
 	return 0;
 }
 
-void
+static void
 make_ref(mp, head)
 struct message *mp;
 struct header *head;
@@ -225,7 +244,7 @@ struct header *head;
 		reflen += strlen(oldref) + 2;
 	if (oldmsgid)
 		reflen += strlen(oldmsgid);
-	newref = (char*)smalloc(reflen);
+	newref = ac_alloc(reflen);
 	if (oldref) {
 		strcpy(newref, oldref);
 		if (oldmsgid) {
@@ -235,7 +254,7 @@ struct header *head;
 	} else if (oldmsgid)
 		strcpy(newref, oldmsgid);
 	n = extract(newref, GREF);
-	free(newref);
+	ac_free(newref);
 	/*
 	 * Limit the references to 21 entries.
 	 */
@@ -251,76 +270,106 @@ struct header *head;
 	head->h_ref = n;
 }
 
+#ifdef	__STDC__
+static int
+(*respond_or_Respond(int c))(int *, int)
+#else
+static int
+(*respond_or_Respond(c))()
+#endif
+{
+	int opt = 0;
+
+	opt += (value("Replyall") != NULL);
+	opt += (value("flipr") != NULL);
+	return ((opt == 1) ^ (c == 'R')) ? Respond_internal : respond_internal;
+}
+
 int
 respond(v)
 	void *v;
 {
-	int *msgvec = v;
-	if (value("Replyall") == NULL)
-		return (_respond(msgvec));
-	else
-		return (_Respond(msgvec));
+	return (respond_or_Respond('r'))((int *)v, 0);
+}
+
+int
+respondall(v)
+	void *v;
+{
+	return respond_internal((int *)v, 0);
+}
+
+int
+respondsender(v)
+	void *v;
+{
+	return Respond_internal((int *)v, 0);
+}
+
+int
+followup(v)
+	void *v;
+{
+	return (respond_or_Respond('r'))((int *)v, 1);
+}
+
+int
+followupall(v)
+	void *v;
+{
+	return respond_internal((int *)v, 1);
+}
+
+int
+followupsender(v)
+	void *v;
+{
+	return Respond_internal((int *)v, 1);
 }
 
 /*
  * Reply to a list of messages.  Extract each name from the
  * message header and send them off to mail1()
  */
-int
-_respond(msgvec)
+static int
+respond_internal(msgvec, recipient_record)
 	int *msgvec;
 {
 	struct message *mp;
-	char *cp, *rcv, *replyto;
-	char **ap;
-	struct name *np;
+	char *cp, *rcv;
+	struct name *np = NULL;
 	struct header head;
 
 	if (msgvec[1] != 0) {
-		printf("Sorry, can't reply to multiple messages at once\n");
+		printf(catgets(catd, CATSET, 37,
+			"Sorry, can't reply to multiple messages at once\n"));
 		return(1);
 	}
 	mp = &message[msgvec[0] - 1];
 	touch(mp);
-	dot = mp;
-	if ((rcv = skin(hfield("from", mp))) == NULL)
-		rcv = skin(nameof(mp, 1));
-	if ((replyto = skin(hfield("reply-to", mp))) != NULL)
-		np = extract(replyto, GTO);
-	else if ((cp = skin(hfield("to", mp))) != NULL)
-		np = extract(cp, GTO);
-	else
-		np = NIL;
+	setdot(mp);
+	if ((rcv = skin(hfield("reply-to", mp))) == NULL)
+		if ((rcv = skin(hfield("from", mp))) == NULL)
+			rcv = skin(nameof(mp, 1));
+	if (rcv != NULL)
+		np = extract(rcv, GTO);
+	if ((cp = skin(hfield("to", mp))) != NULL)
+		np = cat(np, extract(cp, GTO));
 	np = elide(np);
 	/*
 	 * Delete my name from the reply list,
 	 * and with it, all my alternate names.
 	 */
-	np = delname(np, myname);
-	if (altnames)
-		for (ap = altnames; *ap; ap++)
-			np = delname(np, *ap);
-	if ((cp = skin(value("from"))) != NULL)
-		np = delname(np, cp);
-	if ((cp = skin(value("replyto"))) != NULL)
-		np = delname(np, cp);
-	if (np != NIL && replyto == NULL)
-		np = cat(np, extract(rcv, GTO));
-	else if (np == NIL) {
-		if (replyto != NULL)
-			printf("Empty reply-to field -- replying to author\n");
+	np = delete_alternates(np);
+	if (np == NIL)
 		np = extract(rcv, GTO);
-	}
 	head.h_to = np;
 	if ((head.h_subject = hfield("subject", mp)) == NULL)
 		head.h_subject = hfield("subj", mp);
 	head.h_subject = reedit(head.h_subject);
-	if (replyto == NULL && (cp = skin(hfield("cc", mp))) != NULL) {
+	if ((cp = skin(hfield("cc", mp))) != NULL) {
 		np = elide(extract(cp, GCC));
-		np = delname(np, myname);
-		if (altnames != 0)
-			for (ap = altnames; *ap; ap++)
-				np = delname(np, *ap);
+		np = delete_alternates(np);
 		head.h_cc = np;
 	} else
 		head.h_cc = NIL;
@@ -328,7 +377,7 @@ _respond(msgvec)
 	make_ref(mp, &head);
 	head.h_attach = NULL;
 	head.h_smopts = NIL;
-	mail1(&head, 1, mp, NULL);
+	mail1(&head, 1, mp, NULL, recipient_record, 0);
 	return(0);
 }
 
@@ -336,7 +385,7 @@ _respond(msgvec)
  * Modify the subject we are replying to to begin with Re: if
  * it does not already.
  */
-char *
+static char *
 reedit(subj)
 	char *subj;
 {
@@ -371,7 +420,8 @@ preserve(v)
 	int *ip, mesg;
 
 	if (edit) {
-		printf("Cannot \"preserve\" in edit mode\n");
+		printf(catgets(catd, CATSET, 39,
+				"Cannot \"preserve\" in edit mode\n"));
 		return(1);
 	}
 	for (ip = msgvec; *ip != 0; ip++) {
@@ -379,7 +429,7 @@ preserve(v)
 		mp = &message[mesg-1];
 		mp->m_flag |= MPRESERVE;
 		mp->m_flag &= ~MBOX;
-		dot = mp;
+		setdot(mp);
 	}
 	return(0);
 }
@@ -395,7 +445,7 @@ unread(v)
 	int *ip;
 
 	for (ip = msgvec; *ip != 0; ip++) {
-		dot = &message[*ip-1];
+		setdot(&message[*ip-1]);
 		dot->m_flag &= ~(MREAD|MTOUCH);
 		dot->m_flag |= MSTATUS;
 	}
@@ -416,8 +466,8 @@ messize(v)
 	for (ip = msgvec; *ip != 0; ip++) {
 		mesg = *ip;
 		mp = &message[mesg-1];
-		printf("%d: %d/%u\n", mesg, mp->m_lines,
-				(unsigned int)mp->m_size);
+		printf(catgets(catd, CATSET, 40, "%d: %d/%u\n"),
+				mesg, mp->m_lines, (unsigned int)mp->m_size);
 	}
 	return(0);
 }
@@ -437,9 +487,18 @@ rexit(v)
 	/*NOTREACHED*/
 }
 
+static sigjmp_buf	pipejmp;
+
+/*ARGSUSED*/
+static RETSIGTYPE
+onpipe(signo)
+{
+	siglongjmp(pipejmp, 1);
+}
+
 /*
  * Set or display a variable value.  Syntax is similar to that
- * of csh.
+ * of sh.
  */
 int
 set(v)
@@ -448,9 +507,14 @@ set(v)
 	char **arglist = v;
 	struct var *vp;
 	char *cp, *cp2;
-	char varbuf[BUFSIZ], **ap, **p;
+	char **ap, **p;
 	int errs, h, s;
+	FILE *obuf = stdout;
+	int bsdcompat = (value("bsdcompat") != NULL);
 
+	(void)&cp;
+	(void)&ap;
+	(void)&obuf;
 	if (*arglist == NULL) {
 		for (h = 0, s = 1; h < HSHSIZE; h++)
 			for (vp = variables[h]; vp != NOVAR; vp = vp->v_link)
@@ -462,12 +526,42 @@ set(v)
 				*p++ = vp->v_name;
 		*p = NULL;
 		sort(ap);
-		for (p = ap; *p != NULL; p++)
-			printf("%s\t%s\n", *p, value(*p));
+		if (is_a_tty[0] && is_a_tty[1] && (cp = value("crt")) != NULL) {
+			if (s > (*cp == '\0' ? screensize() : atoi(cp)) + 3) {
+				cp = get_pager();
+				if (sigsetjmp(pipejmp, 1))
+					goto endpipe;
+				if ((obuf = Popen(cp, "w", NULL, 1)) == NULL) {
+					perror(cp);
+					obuf = stdout;
+				} else
+					safe_signal(SIGPIPE, onpipe);
+			}
+		}
+		for (p = ap; *p != NULL; p++) {
+			if (bsdcompat)
+				fprintf(obuf, "%s\t%s\n", *p, value(*p));
+			else {
+				if ((cp = value(*p)) != NULL && *cp)
+					fprintf(obuf, "%s=\"%s\"\n",
+							*p, value(*p));
+				else
+					fprintf(obuf, "%s\n", *p);
+			}
+		}
+endpipe:
+		if (obuf != stdout) {
+			safe_signal(SIGPIPE, SIG_IGN);
+			Pclose(obuf);
+			safe_signal(SIGPIPE, SIG_DFL);
+		}
 		return(0);
 	}
 	errs = 0;
 	for (ap = arglist; *ap != NULL; ap++) {
+		char *varbuf;
+
+		varbuf = ac_alloc(strlen(*ap) + 1);
 		cp = *ap;
 		cp2 = varbuf;
 		while (*cp != '=' && *cp != '\0')
@@ -478,11 +572,17 @@ set(v)
 		else
 			cp++;
 		if (equal(varbuf, "")) {
-			printf("Non-null variable name required\n");
+			printf(catgets(catd, CATSET, 41,
+					"Non-null variable name required\n"));
 			errs++;
+			ac_free(varbuf);
 			continue;
 		}
-		assign(varbuf, cp);
+		if (varbuf[0] == 'n' && varbuf[1] == 'o')
+			errs += unset_internal(&varbuf[2]);
+		else
+			assign(varbuf, cp);
+		ac_free(varbuf);
 	}
 	return(errs);
 }
@@ -494,35 +594,12 @@ int
 unset(v)
 	void *v;
 {
-	char **arglist = v;
-	struct var *vp, *vp2;
-	int errs, h;
+	int errs;
 	char **ap;
 
 	errs = 0;
-	for (ap = arglist; *ap != NULL; ap++) {
-		if ((vp2 = lookup(*ap)) == NOVAR) {
-			if (!sourcing) {
-				printf("\"%s\": undefined variable\n", *ap);
-				errs++;
-			}
-			continue;
-		}
-		h = hash(*ap);
-		if (vp2 == variables[h]) {
-			variables[h] = variables[h]->v_link;
-			vfree(vp2->v_name);
-			vfree(vp2->v_value);
-			free((char *)vp2);
-			continue;
-		}
-		for (vp = variables[h]; vp->v_link != vp2; vp = vp->v_link)
-			;
-		vp->v_link = vp2->v_link;
-		vfree(vp2->v_name);
-		vfree(vp2->v_value);
-		free((char *) vp2);
-	}
+	for (ap = (char **)v; *ap != NULL; ap++)
+		errs += unset_internal(*ap);
 	return(errs);
 }
 
@@ -562,7 +639,7 @@ group(v)
 	gname = *argv;
 	h = hash(gname);
 	if ((gh = findgroup(gname)) == NOGRP) {
-		gh = (struct grouphead *) calloc(sizeof *gh, 1);
+		gh = (struct grouphead *)scalloc(1, sizeof *gh);
 		gh->g_name = vcopy(gname);
 		gh->g_list = NOGE;
 		gh->g_link = groups[h];
@@ -576,7 +653,7 @@ group(v)
 	 */
 
 	for (ap = argv+1; *ap != NULL; ap++) {
-		gp = (struct group *) calloc(sizeof *gp, 1);
+		gp = (struct group *)scalloc(1, sizeof *gp);
 		gp->ge_name = vcopy(*ap);
 		gp->ge_link = gh->g_list;
 		gh->g_list = gp;
@@ -584,11 +661,31 @@ group(v)
 	return(0);
 }
 
+/*
+ * Delete the passed groups.
+ */
+int
+ungroup(v)
+	void *v;
+{
+	char **argv = v;
+
+	if (*argv == NULL) {
+		printf(catgets(catd, CATSET, 209,
+				"Must specify alias or group to remove\n"));
+		return 1;
+	}
+	do
+		remove_group(*argv);
+	while (*++argv != NULL);
+	return 0;
+}
+
 /*
  * Sort the passed string vecotor into ascending dictionary
  * order.
  */
-void
+static void
 sort(list)
 	char **list;
 {
@@ -627,12 +724,12 @@ file(v)
 		newfileinfo();
 		return 0;
 	}
-	i = setfile(*argv);
+	i = setfile(*argv, 0);
 	if (i < 0)
 		return 1;
 	if (i > 0 && value("emptystart") == NULL)
 		return 1;
-	announce();
+	announce(value("bsdcompat") != NULL);
 	return 0;
 }
 
@@ -651,11 +748,11 @@ echo(v)
 		cp = *ap;
 		if ((cp = expand(cp)) != NULL) {
 			if (ap != argv)
-				putchar(' ');
+				sputc(' ', stdout);
 			printf("%s", cp);
 		}
 	}
-	putchar('\n');
+	sputc('\n', stdout);
 	return 0;
 }
 
@@ -663,11 +760,14 @@ int
 Respond(v)
 	void *v;
 {
-	int *msgvec = v;
-	if (value("Replyall") == NULL)
-		return (_Respond(msgvec));
-	else
-		return (_respond(msgvec));
+	return (respond_or_Respond('R'))((int *)v, 0);
+}
+
+int
+Followup(v)
+	void *v;
+{
+	return (respond_or_Respond('R'))((int *)v, 1);
 }
 
 /*
@@ -675,8 +775,8 @@ Respond(v)
  * and not messing around with the To: and Cc: lists as in normal
  * reply.
  */
-int
-_Respond(msgvec)
+static int
+Respond_internal(msgvec, recipient_record)
 	int msgvec[];
 {
 	struct header head;
@@ -688,9 +788,10 @@ _Respond(msgvec)
 	for (ap = msgvec; *ap != 0; ap++) {
 		mp = &message[*ap - 1];
 		touch(mp);
-		dot = mp;
-		if ((cp = skin(hfield("from", mp))) == NULL)
-			cp = skin(nameof(mp, 2));
+		setdot(mp);
+		if ((cp = skin(hfield("reply-to", mp))) == NULL)
+			if ((cp = skin(hfield("from", mp))) == NULL)
+				cp = skin(nameof(mp, 2));
 		head.h_to = cat(head.h_to, extract(cp, GTO));
 	}
 	if (head.h_to == NIL)
@@ -704,7 +805,7 @@ _Respond(msgvec)
 	make_ref(mp, &head);
 	head.h_attach = NULL;
 	head.h_smopts = NIL;
-	mail1(&head, 1, mp, NULL);
+	mail1(&head, 1, mp, NULL, recipient_record, 0);
 	return 0;
 }
 
@@ -720,7 +821,7 @@ ifcmd(v)
 	char *cp;
 
 	if (cond != CANY) {
-		printf("Illegal nested \"if\"\n");
+		printf(catgets(catd, CATSET, 42, "Illegal nested \"if\"\n"));
 		return(1);
 	}
 	cond = CANY;
@@ -734,8 +835,13 @@ ifcmd(v)
 		cond = CSEND;
 		break;
 
+	case 't': case 'T':
+		cond = CTERM;
+		break;
+
 	default:
-		printf("Unrecognized if-keyword: \"%s\"\n", cp);
+		printf(catgets(catd, CATSET, 43,
+				"Unrecognized if-keyword: \"%s\"\n"), cp);
 		return(1);
 	}
 	return(0);
@@ -753,7 +859,8 @@ elsecmd(v)
 
 	switch (cond) {
 	case CANY:
-		printf("\"Else\" without matching \"if\"\n");
+		printf(catgets(catd, CATSET, 44,
+				"\"Else\" without matching \"if\"\n"));
 		return(1);
 
 	case CSEND:
@@ -764,8 +871,13 @@ elsecmd(v)
 		cond = CSEND;
 		break;
 
+	case CTERM:
+		cond = CNONTERM;
+		break;
+
 	default:
-		printf("Mail's idea of conditions is screwed up\n");
+		printf(catgets(catd, CATSET, 45,
+				"Mail's idea of conditions is screwed up\n"));
 		cond = CANY;
 		break;
 	}
@@ -782,7 +894,8 @@ endifcmd(v)
 {
 
 	if (cond == CANY) {
-		printf("\"Endif\" without matching \"if\"\n");
+		printf(catgets(catd, CATSET, 46,
+				"\"Endif\" without matching \"if\"\n"));
 		return(1);
 	}
 	cond = CANY;
@@ -811,9 +924,9 @@ alternates(v)
 	}
 	if (altnames != 0)
 		free((char *) altnames);
-	altnames = (char **) calloc((unsigned) c, sizeof (char *));
+	altnames = (char **)scalloc(c, sizeof (char *));
 	for (ap = namelist, ap2 = altnames; *ap; ap++, ap2++) {
-		cp = (char *) calloc((unsigned) strlen(*ap) + 1, sizeof (char));
+		cp = (char *)scalloc(strlen(*ap) + 1, sizeof (char));
 		strcpy(cp, *ap);
 		*ap2 = cp;
 	}
@@ -824,7 +937,7 @@ alternates(v)
 /*
  * Do the real work of forwarding.
  */
-int
+static int
 forward1(v, add_resent)
 void *v;
 {
@@ -836,15 +949,16 @@ void *v;
 	str = (char *)v;
 	/*LINTED*/
 	msgvec = (int *)salloc((msgcount + 2) * sizeof *msgvec);
-	name = getcmd(str, &f);
+	name = laststring(str, &f, 1);
 	if (name == NULL) {
-		puts("No recipient specified.");
+		puts(catgets(catd, CATSET, 47, "No recipient specified."));
 		return 1;
 	}
 	if (!f) {
 		*msgvec = first(0, MMNORM);
 		if (*msgvec == 0) {
-			puts("No applicable messages.");
+			puts(catgets(catd, CATSET, 48,
+					"No applicable messages."));
 			return 1;
 		}
 		msgvec[1] = 0;
@@ -880,3 +994,121 @@ void *v;
 {
 	return forward1(v, 0);
 }
+
+/*
+ * 'newmail' or 'inc' command: Check for new mail without writing old
+ * mail back.
+ */
+/*ARGSUSED*/
+int
+newmail(v)
+void *v;
+{
+	int omsgcount = msgcount;
+	int val;
+
+	if ((val = setfile(mailname, 1)) == 0) {
+		newfileinfo();
+		setdot(&message[omsgcount]);
+	}
+	return val;
+}
+
+static void
+list_shortcuts()
+{
+	struct shortcut *s;
+
+	for (s = shortcuts; s; s = s->sh_next)
+		printf("%s=%s\n", s->sh_short, s->sh_long);
+}
+
+int
+shortcut(v)
+	void *v;
+{
+	char **args = (char **)v;
+	struct shortcut *s;
+
+	if (args[0] == NULL) {
+		list_shortcuts();
+		return 0;
+	}
+	if (args[1] == NULL) {
+		fprintf(stderr, catgets(catd, CATSET, 220,
+				"expansion name for shortcut missing\n"));
+		return 1;
+	}
+	if (args[2] != NULL) {
+		fprintf(stderr, catgets(catd, CATSET, 221,
+				"too many arguments\n"));
+		return 1;
+	}
+	if ((s = get_shortcut(args[0])) != NULL) {
+		free(s->sh_long);
+		s->sh_long = sstrdup(args[1]);
+	} else {
+		s = scalloc(1, sizeof *s);
+		s->sh_short = sstrdup(args[0]);
+		s->sh_long = sstrdup(args[1]);
+		s->sh_next = shortcuts;
+		shortcuts = s;
+	}
+	return 0;
+}
+
+struct shortcut *
+get_shortcut(str)
+	const char *str;
+{
+	struct shortcut *s;
+
+	for (s = shortcuts; s; s = s->sh_next)
+		if (strcmp(str, s->sh_short) == 0)
+			break;
+	return s;
+}
+
+static enum okay
+delete_shortcut(str)
+	const char *str;
+{
+	struct shortcut *sp, *sq;
+
+	for (sp = shortcuts, sq = NULL; sp; sq = sp, sp = sp->sh_next) {
+		if (strcmp(sp->sh_short, str) == 0) {
+			free(sp->sh_short);
+			free(sp->sh_long);
+			if (sq)
+				sq->sh_next = sp->sh_next;
+			if (sp == shortcuts)
+				shortcuts = sp->sh_next;
+			free(sp);
+			return OKAY;
+		}
+	}
+	return STOP;
+}
+
+int
+unshortcut(v)
+	void *v;
+{
+	char **args = (char **)v;
+	int errs = 0;
+
+	if (args[0] == NULL) {
+		fprintf(stderr, catgets(catd, CATSET, 222,
+				"need shortcut names to remove\n"));
+		return 1;
+	}
+	while (*args != NULL) {
+		if (delete_shortcut(*args) != OKAY) {
+			errs = 1;
+			fprintf(stderr, catgets(catd, CATSET, 223,
+				"%s: no such shortcut\n"), *args);
+		}
+		args++;
+	}
+	return errs;
+}
