@@ -1,3 +1,8 @@
+/*
+ * Nail - a mail user agent derived from Berkeley Mail.
+ *
+ * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ */
 /*
  * Copyright (c) 1980, 1993
  *	The Regents of the University of California.  All rights reserved.
@@ -33,7 +38,7 @@
 
 #ifndef lint
 #ifdef	DOSCCS
-static char sccsid[] = "@(#)list.c	1.6 (gritter) 5/23/02";
+static char sccsid[] = "@(#)list.c	2.11 (gritter) 1/10/03";
 #endif
 #endif /* not lint */
 
@@ -46,6 +51,26 @@ static char sccsid[] = "@(#)list.c	1.6 (gritter) 5/23/02";
  *
  * Message list handling.
  */
+static char	**add_to_namelist __P((char ***, size_t *, char **, char *));
+static int	markall __P((char [], int));
+static int	evalcol __P((int));
+static int	check __P((int, int));
+static int	scan __P((char **));
+static void	regret __P((int));
+static void	scaninit __P((void));
+static int	matchsender __P((char *, int, int));
+static int	matchmid __P((char *, int));
+static int	matchsubj __P((char *, int));
+static void	mark __P((int));
+static void	unmark __P((int));
+static int	metamess __P((int, int));
+
+static int	lexnumber;		/* Number of TNUMBER from scan() */
+static char	lexstring[STRINGLEN];	/* String from TSTRING, scan() */
+static int	regretp;		/* Pointer to TOS of regret tokens */
+static int	regretstack[REGDEP];	/* Stack of regretted tokens */
+static char	*string_stack[REGDEP];	/* Stack of regretted strings */
+static int	numberstack[REGDEP];	/* Stack of regretted numbers */
 
 /*
  * Convert the user string of message numbers and
@@ -96,7 +121,7 @@ getmsglist(buf, vector, flags)
  * the colon and gives the corresponding modifier bit.
  */
 
-struct coltab {
+static struct coltab {
 	char	co_char;		/* What to find past : */
 	int	co_bit;			/* Associated modifier bit */
 	int	co_mask;		/* m_status bits to mask */
@@ -112,16 +137,39 @@ struct coltab {
 
 static	int	lastcolmod;
 
-int
+static char **
+add_to_namelist(namelist, nmlsize, np, string)
+	char ***namelist;
+	size_t *nmlsize;
+	char **np;
+	char *string;
+{
+	size_t idx;
+
+	if ((idx = np - *namelist) >= *nmlsize) {
+		*namelist = srealloc(*namelist, (*nmlsize += 8) * sizeof *np);
+		np = &(*namelist)[idx];
+	}
+	*np++ = string;
+	return np;
+}
+
+#define	markall_ret(i)		{ \
+					retval = i; \
+					goto out; \
+				}
+
+static int
 markall(buf, f)
 	char buf[];
 	int f;
 {
-	char **np;
-	int i;
+	char **np, **nq;
+	int i, retval;
 	struct message *mp;
-	char *namelist[NMLSIZE], *bufp;
+	char **namelist, *bufp, *id = NULL, *cp;
 	int tok, beg, mc, star, other, valdot, colmod, colresult;
+	size_t nmlsize;
 
 	valdot = dot - &message[0] + 1;
 	colmod = 0;
@@ -129,6 +177,7 @@ markall(buf, f)
 		unmark(i);
 	bufp = buf;
 	mc = 0;
+	namelist = smalloc((nmlsize = 8) * sizeof *namelist);
 	np = &namelist[0];
 	scaninit();
 	tok = scan(&bufp);
@@ -140,14 +189,15 @@ markall(buf, f)
 		case TNUMBER:
 number:
 			if (star) {
-				printf("No numbers mixed with *\n");
-				return(-1);
+				printf(catgets(catd, CATSET, 112,
+					"No numbers mixed with *\n"));
+				markall_ret(-1)
 			}
 			mc++;
 			other++;
 			if (beg != 0) {
 				if (check(lexnumber, f))
-					return(-1);
+					markall_ret(-1)
 				for (i = beg; i <= lexnumber; i++)
 					if (f == MDELETED || (message[i - 1].m_flag & MDELETED) == 0)
 						mark(i);
@@ -156,7 +206,7 @@ number:
 			}
 			beg = lexnumber;
 			if (check(beg, f))
-				return(-1);
+				markall_ret(-1)
 			tok = scan(&bufp);
 			regret(tok);
 			if (tok != TDASH) {
@@ -167,15 +217,17 @@ number:
 
 		case TPLUS:
 			if (beg != 0) {
-				printf("Non-numeric second argument\n");
-				return(-1);
+				printf(catgets(catd, CATSET, 113,
+					"Non-numeric second argument\n"));
+				markall_ret(-1)
 			}
 			i = valdot;
 			do {
 				i++;
 				if (i > msgcount) {
-					printf("Referencing beyond EOF\n");
-					return(-1);
+					printf(catgets(catd, CATSET, 114,
+						"Referencing beyond EOF\n"));
+					markall_ret(-1)
 				}
 			} while ((message[i - 1].m_flag & MDELETED) != f);
 			mark(i);
@@ -187,8 +239,10 @@ number:
 				do {
 					i--;
 					if (i <= 0) {
-						printf("Referencing before 1\n");
-						return(-1);
+						printf(catgets(catd, CATSET,
+							115,
+						"Referencing before 1\n"));
+						markall_ret(-1)
 					}
 				} while ((message[i - 1].m_flag & MDELETED) != f);
 				mark(i);
@@ -197,46 +251,71 @@ number:
 
 		case TSTRING:
 			if (beg != 0) {
-				printf("Non-numeric second argument\n");
-				return(-1);
+				printf(catgets(catd, CATSET, 116,
+					"Non-numeric second argument\n"));
+				markall_ret(-1)
 			}
 			other++;
 			if (lexstring[0] == ':') {
 				colresult = evalcol(lexstring[1]);
 				if (colresult == 0) {
-					printf("Unknown colon modifier \"%s\"\n",
+					printf(catgets(catd, CATSET, 117,
+					"Unknown colon modifier \"%s\"\n"),
 					    lexstring);
-					return(-1);
+					markall_ret(-1)
 				}
 				colmod |= colresult;
 			}
 			else
-				*np++ = savestr(lexstring);
+				np = add_to_namelist(&namelist, &nmlsize,
+						np, savestr(lexstring));
 			break;
 
 		case TDOLLAR:
 		case TUP:
 		case TDOT:
+		case TSEMI:
 			lexnumber = metamess(lexstring[0], f);
 			if (lexnumber == -1)
-				return(-1);
+				markall_ret(-1)
 			goto number;
 
 		case TSTAR:
 			if (other) {
-				printf("Can't mix \"*\" with anything\n");
-				return(-1);
+				printf(catgets(catd, CATSET, 118,
+					"Can't mix \"*\" with anything\n"));
+				markall_ret(-1)
 			}
 			star++;
 			break;
 
+		case TCOMMA:
+			if ((cp = hfield("in-reply-to", dot)) != NULL)
+				id = savestr(cp);
+			else if ((cp = hfield("references", dot)) != NULL) {
+				struct name *n;
+
+				if ((n = extract(cp, GREF)) != NULL) {
+					while (n->n_flink != NULL)
+						n = n->n_flink;
+					id = savestr(n->n_name);
+				}
+			}
+			if (id == NULL) {
+				printf(catgets(catd, CATSET, 227,
+		"Cannot determine parent Message-ID of the current message\n"));
+				markall_ret(-1)
+			}
+			break;
+
 		case TERROR:
-			return -1;
+			markall_ret(-1)
 		}
 		tok = scan(&bufp);
 	}
 	lastcolmod = colmod;
-	*np = NULL;
+	np = add_to_namelist(&namelist, &nmlsize, np, NULL);
+	np--;
 	mc = 0;
 	if (star) {
 		for (i = 0; i < msgcount; i++)
@@ -245,10 +324,11 @@ number:
 				mc++;
 			}
 		if (mc == 0) {
-			printf("No applicable messages.\n");
-			return(-1);
+			printf(catgets(catd, CATSET, 119,
+					"No applicable messages.\n"));
+			markall_ret(-1)
 		}
-		return(0);
+		markall_ret(0)
 	}
 
 	/*
@@ -257,7 +337,7 @@ number:
 	 * if any user names were given.
 	 */
 
-	if ((np > namelist || colmod != 0) && mc == 0)
+	if ((np > namelist || colmod != 0 || id) && mc == 0)
 		for (i = 1; i <= msgcount; i++)
 			if ((message[i-1].m_flag & MDELETED) == f)
 				mark(i);
@@ -267,21 +347,30 @@ number:
 	 * messages whose senders were not requested.
 	 */
 
-	if (np > namelist) {
+	if (np > namelist || id) {
+		int	allnet = value("allnet") != NULL;
+
 		for (i = 1; i <= msgcount; i++) {
-			for (mc = 0, np = &namelist[0]; *np != NULL; np++)
-				if (**np == '/') {
-					if (matchsubj(*np, i)) {
-						mc++;
-						break;
+			mc = 0;
+			if (np > namelist) {
+				for (nq = &namelist[0]; *nq != NULL; nq++) {
+					if (**nq == '/') {
+						if (matchsubj(*nq, i)) {
+							mc++;
+							break;
+						}
 					}
-				}
-				else {
-					if (matchsender(*np, i)) {
-						mc++;
-						break;
+					else {
+						if (matchsender(*nq, i,
+								allnet)) {
+							mc++;
+							break;
+						}
 					}
 				}
+			}
+			if (mc == 0 && id && matchmid(id, i))
+				mc++;
 			if (mc == 0)
 				unmark(i);
 		}
@@ -297,12 +386,19 @@ number:
 				break;
 			}
 		if (mc == 0) {
-			printf("No applicable messages from {%s",
-				namelist[0]);
-			for (np = &namelist[1]; *np != NULL; np++)
-				printf(", %s", *np);
-			printf("}\n");
-			return(-1);
+			if (np > namelist) {
+				printf(catgets(catd, CATSET, 120,
+					"No applicable messages from {%s"),
+					namelist[0]);
+				for (nq = &namelist[1]; *nq != NULL; nq++)
+					printf(catgets(catd, CATSET, 121,
+								", %s"), *nq);
+				printf(catgets(catd, CATSET, 122, "}\n"));
+			} else if (id) {
+				printf(catgets(catd, CATSET, 227,
+					"Parent message not found\n"));
+			}
+			markall_ret(-1)
 		}
 	}
 
@@ -329,22 +425,26 @@ number:
 		if (mp >= &message[msgcount]) {
 			struct coltab *colp;
 
-			printf("No messages satisfy");
+			printf(catgets(catd, CATSET, 123,
+						"No messages satisfy"));
 			for (colp = &coltab[0]; colp->co_char; colp++)
 				if (colp->co_bit & colmod)
 					printf(" :%c", colp->co_char);
 			printf("\n");
-			return(-1);
+			markall_ret(-1)
 		}
 	}
-	return(0);
+	markall_ret(0)
+out:
+	free(namelist);
+	return retval;
 }
 
 /*
  * Turn the character after a colon modifier into a bit
  * value.
  */
-int
+static int
 evalcol(col)
 	int col;
 {
@@ -363,19 +463,21 @@ evalcol(col)
  * If f is MDELETED, then either kind will do.  Otherwise, the message
  * has to be undeleted.
  */
-int
+static int
 check(mesg, f)
 	int mesg, f;
 {
 	struct message *mp;
 
 	if (mesg < 1 || mesg > msgcount) {
-		printf("%d: Invalid message number\n", mesg);
+		printf(catgets(catd, CATSET, 124,
+			"%d: Invalid message number\n"), mesg);
 		return(-1);
 	}
 	mp = &message[mesg-1];
 	if (f != MDELETED && (mp->m_flag & MDELETED) != 0) {
-		printf("%d: Inappropriate message\n", mesg);
+		printf(catgets(catd, CATSET, 125,
+			"%d: Inappropriate message\n"), mesg);
 		return(-1);
 	}
 	return(0);
@@ -386,25 +488,26 @@ check(mesg, f)
  * for a RAWLIST.
  */
 int
-getrawlist(line, argv, argc)
+getrawlist(line, linesize, argv, argc)
 	char line[];
+	size_t linesize;
 	char **argv;
 	int  argc;
 {
 	char c, *cp, *cp2, quotec;
 	int argn;
-	char linebuf[BUFSIZ];
+	char *linebuf;
 
 	argn = 0;
 	cp = line;
+	linebuf = ac_alloc(linesize + 1);
 	for (;;) {
-		for (; *cp == ' ' || *cp == '\t'; cp++)
-			;
+		for (; blankchar(*cp & 0377); cp++);
 		if (*cp == '\0')
 			break;
 		if (argn >= argc - 1) {
-			printf(
-			"Too many elements in the list; excess discarded.\n");
+			printf(catgets(catd, CATSET, 126,
+			"Too many elements in the list; excess discarded.\n"));
 			break;
 		}
 		cp2 = linebuf;
@@ -466,7 +569,7 @@ getrawlist(line, argv, argc)
 					*cp2++ = c;
 			} else if (c == '"' || c == '\'')
 				quotec = c;
-			else if (c == ' ' || c == '\t')
+			else if (blankchar(c & 0377))
 				break;
 			else
 				*cp2++ = c;
@@ -475,6 +578,7 @@ getrawlist(line, argv, argc)
 		argv[argn++] = savestr(linebuf);
 	}
 	argv[argn] = NULL;
+	ac_free(linebuf);
 	return argn;
 }
 
@@ -485,9 +589,9 @@ getrawlist(line, argv, argc)
  * appropriate.  In any event, store the scanned `thing' in lexstring.
  */
 
-struct lex {
+static struct lex {
 	char	l_char;
-	char	l_token;
+	enum ltoken	l_token;
 } singles[] = {
 	{ '$',	TDOLLAR },
 	{ '.',	TDOT },
@@ -497,10 +601,12 @@ struct lex {
 	{ '+',	TPLUS },
 	{ '(',	TOPEN },
 	{ ')',	TCLOSE },
+	{ ',',	TCOMMA },
+	{ ';',	TSEMI },
 	{ 0,	0 }
 };
 
-int
+static int
 scan(sp)
 	char **sp;
 {
@@ -523,7 +629,7 @@ scan(sp)
 	 * strip away leading white space.
 	 */
 
-	while (c == ' ' || c == '\t')
+	while (blankchar(c))
 		c = *cp++;
 
 	/*
@@ -542,9 +648,9 @@ scan(sp)
 	 * Return TNUMBER when done.
 	 */
 
-	if (isdigit(c)) {
+	if (digitchar(c)) {
 		lexnumber = 0;
-		while (isdigit(c)) {
+		while (digitchar(c)) {
 			lexnumber = lexnumber*10 + c - '0';
 			*cp2++ = c;
 			c = *cp++;
@@ -585,14 +691,15 @@ scan(sp)
 			cp++;
 			break;
 		}
-		if (quotec == 0 && (c == ' ' || c == '\t'))
+		if (quotec == 0 && blankchar(c))
 			break;
 		if (cp2 - lexstring < STRINGLEN-1)
 			*cp2++ = c;
 		c = *cp++;
 	}
 	if (quotec && c == 0) {
-		fprintf(stderr, "Missing %c\n", quotec);
+		fprintf(stderr, catgets(catd, CATSET, 127,
+				"Missing %c\n"), quotec);
 		return TERROR;
 	}
 	*sp = --cp;
@@ -603,12 +710,12 @@ scan(sp)
 /*
  * Unscan the named token by pushing it onto the regret stack.
  */
-void
+static void
 regret(token)
 	int token;
 {
 	if (++regretp >= REGDEP)
-		panic("Too many regrets");
+		panic(catgets(catd, CATSET, 128, "Too many regrets"));
 	regretstack[regretp] = token;
 	lexstring[STRINGLEN-1] = '\0';
 	string_stack[regretp] = savestr(lexstring);
@@ -618,7 +725,7 @@ regret(token)
 /*
  * Reset all the scanner global variables.
  */
-void
+static void
 scaninit()
 {
 	regretp = -1;
@@ -651,12 +758,37 @@ first(f, m)
  * See if the passed name sent the passed message number.  Return true
  * if so.
  */
-int
-matchsender(str, mesg)
+static int
+matchsender(str, mesg, allnet)
 	char *str;
+	int mesg, allnet;
+{
+	if (allnet) {
+		char *cp = nameof(&message[mesg - 1], 0);
+
+		do {
+			if ((*cp == '@' || *cp == '\0') &&
+					(*str == '@' || *str == '\0'))
+				return 1;
+			if (*cp != *str)
+				break;
+		} while (cp++, *str++ != '\0');
+		return 0;
+	}
+	return !strcmp(str, nameof(&message[mesg - 1], 0));
+}
+
+static int
+matchmid(id, mesg)
+	char *id;
 	int mesg;
 {
-	return !strcmp(str, nameof(&message[mesg - 1], 0));
+	char *cp;
+
+	if ((cp = hfield("message-id", &message[mesg - 1])) != NULL &&
+			strcmp(cp, id) == 0)
+		return 1;
+	return 0;
 }
 
 /*
@@ -667,21 +799,23 @@ matchsender(str, mesg)
  * previous search string.
  */
 
-char lastscan[128];
-int
+static char lastscan[128];
+
+static int
 matchsubj(str, mesg)
 	char *str;
 	int mesg;
 {
 	struct message *mp;
 	char *cp, *cp2, *backup;
+	struct str in, out;
 
 	str++;
 	if (strlen(str) == 0) {
 		str = lastscan;
 	} else {
-		strncpy(lastscan, str, 128);
-		lastscan[127]='\0';
+		strncpy(lastscan, str, sizeof lastscan);
+		lastscan[sizeof lastscan - 1]='\0';
 	}
 	mp = &message[mesg-1];
 	
@@ -700,22 +834,62 @@ matchsubj(str, mesg)
 	}
 	if (cp2 == NULL)
 		return(0);
+	in.s = cp2;
+	in.l = strlen(cp2);
+	mime_fromhdr(&in, &out, TD_ICONV);
+	cp2 = out.s;
 	backup = cp2;
 	while (*cp2) {
-		if (*cp == 0)
+		if (*cp == 0) {
+			free(out.s);
 			return(1);
-		if (aux_raise(*cp++) != aux_raise(*cp2++)) {
-			cp2 = ++backup;
-			cp = str;
+		}
+#if defined (HAVE_MBTOWC) && defined (HAVE_TOWUPPER)
+		if (mb_cur_max > 1) {
+			wchar_t c, c2;
+			int sz;
+
+			if ((sz = mbtowc(&c, cp, mb_cur_max)) < 0)
+				goto singlebyte;
+			cp += sz;
+			if ((sz = mbtowc(&c2, cp2, mb_cur_max)) < 0)
+				goto singlebyte;
+			cp2 += sz;
+			c = towupper(c);
+			c2 = towupper(c2);
+			if (c != c2) {
+				if ((sz = mbtowc(&c, backup, mb_cur_max)) > 0) {
+					backup += sz;
+					cp2 = backup;
+				} else
+					cp2 = ++backup;
+				cp = str;
+			}
+		} else
+#endif	/* HAVE_MBTOWC && HAVE_TOWUPPER */
+		{
+			int c, c2;
+
+	singlebyte:	c = *cp++ & 0377;
+			if (islower(c))
+				c = toupper(c);
+			c2 = *cp2++ & 0377;
+			if (islower(c2))
+				c2 = toupper(c2);
+			if (c != c2) {
+				cp2 = ++backup;
+				cp = str;
+			}
 		}
 	}
+	free(out.s);
 	return(*cp == 0);
 }
 
 /*
  * Mark the named message by setting its mark bit.
  */
-void
+static void
 mark(mesg)
 	int mesg;
 {
@@ -723,14 +897,14 @@ mark(mesg)
 
 	i = mesg;
 	if (i < 1 || i > msgcount)
-		panic("Bad message number to mark");
+		panic(catgets(catd, CATSET, 129, "Bad message number to mark"));
 	message[i-1].m_flag |= MMARK;
 }
 
 /*
  * Unmark the named message.
  */
-void
+static void
 unmark(mesg)
 	int mesg;
 {
@@ -738,14 +912,15 @@ unmark(mesg)
 
 	i = mesg;
 	if (i < 1 || i > msgcount)
-		panic("Bad message number to unmark");
+		panic(catgets(catd, CATSET, 130,
+					"Bad message number to unmark"));
 	message[i-1].m_flag &= ~MMARK;
 }
 
 /*
  * Return the message number corresponding to the passed meta character.
  */
-int
+static int
 metamess(meta, f)
 	int meta, f;
 {
@@ -761,7 +936,7 @@ metamess(meta, f)
 		for (mp = &message[0]; mp < &message[msgcount]; mp++)
 			if ((mp->m_flag & MDELETED) == f)
 				return(mp - &message[0] + 1);
-		printf("No applicable messages\n");
+		printf(catgets(catd, CATSET, 131, "No applicable messages\n"));
 		return(-1);
 
 	case '$':
@@ -771,7 +946,7 @@ metamess(meta, f)
 		for (mp = &message[msgcount-1]; mp >= &message[0]; mp--)
 			if ((mp->m_flag & MDELETED) == f)
 				return(mp - &message[0] + 1);
-		printf("No applicable messages\n");
+		printf(catgets(catd, CATSET, 132, "No applicable messages\n"));
 		return(-1);
 
 	case '.':
@@ -780,13 +955,32 @@ metamess(meta, f)
 		 */
 		m = dot - &message[0] + 1;
 		if ((dot->m_flag & MDELETED) != f) {
-			printf("%d: Inappropriate message\n", m);
+			printf(catgets(catd, CATSET, 133,
+				"%d: Inappropriate message\n"), m);
+			return(-1);
+		}
+		return(m);
+
+	case ';':
+		/*
+		 * Previously current message.
+		 */
+		if (prevdot == NULL) {
+			printf(catgets(catd, CATSET, 228,
+				"No previously current message\n"));
+			return(-1);
+		}
+		m = prevdot - &message[0] + 1;
+		if ((prevdot->m_flag & MDELETED) != f) {
+			printf(catgets(catd, CATSET, 133,
+				"%d: Inappropriate message\n"), m);
 			return(-1);
 		}
 		return(m);
 
 	default:
-		printf("Unknown metachar (%c)\n", c);
+		printf(catgets(catd, CATSET, 134,
+				"Unknown metachar (%c)\n"), c);
 		return(-1);
 	}
 }
