@@ -1,3 +1,8 @@
+/*
+ * Nail - a mail user agent derived from Berkeley Mail.
+ *
+ * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ */
 /*
  * Copyright (c) 2000
  *	Gunnar Ritter.  All rights reserved.
@@ -32,14 +37,11 @@
  */
 
 #ifndef lint
-static char copyright[]
-#ifdef	__GNUC__
-__attribute__ ((unused))
-#endif
-= "@(#) Copyright (c) 2000 Gunnar Ritter. All rights reserved.\n";
 #ifdef	DOSCCS
-static char sccsid[]  = "@(#)mime.c	1.34 (gritter) 6/2/02";
-#endif
+static char copyright[]
+= "@(#) Copyright (c) 2000, 2002 Gunnar Ritter. All rights reserved.\n";
+static char sccsid[]  = "@(#)mime.c	2.12 (gritter) 11/24/02";
+#endif /* DOSCCS */
 #endif /* not lint */
 
 #include "rcv.h"
@@ -64,6 +66,34 @@ static char *mimetypes_world = "/etc/mime.types";
 static char *mimetypes_user = "~/.mime.types";
 char *us_ascii = "us-ascii";
 
+static int	mustquote_body __P((int));
+static int	mustquote_hdr __P((int));
+static int	mustquote_inhdrq __P((int));
+#if defined (HAVE_MBTOWC) && defined (HAVE_ISWPRINT)
+static size_t	xmbstowcs __P((wchar_t *, const char *, size_t));
+#endif
+static char	*getcharset __P((int));
+static int	has_highbit __P((const char *));
+static void	uppercopy __P((char *, const char *));
+static void	stripdash __P((char *));
+#ifdef	HAVE_ICONV
+static size_t	iconv_ft __P((iconv_t, char **, size_t *, char **, size_t *));
+#endif
+static int	is_this_enc __P((const char *, const char *));
+static char	*mime_tline __P((char *, char *));
+static char	*mime_type __P((char *, char *));
+static int	gettextconversion __P((void));
+static enum mimeclean	mime_isclean __P((FILE*));
+static void	invalid_seq __P((int));
+static char	*ctohex __P((int, char *));
+static size_t	mime_write_toqp __P((struct str *, FILE *, int (*)(int)));
+static void	mime_str_toqp __P((struct str *, struct str *, int (*)(int)));
+static void	mime_fromqp __P((struct str *, struct str *, int));
+static size_t	mime_write_tohdr __P((struct str *, FILE *));
+static size_t	mime_write_tohdr_a __P((struct str *, FILE *));
+static size_t	fwrite_td __P((void *, size_t, size_t, FILE *, enum tdflags,
+			char *, size_t));
+
 /*
  * Check if c must be quoted inside a message's body.
  */
@@ -102,7 +132,7 @@ mustquote_inhdrq(c)
 /*
  * A mbstowcs()-alike function that transparently handles invalid sequences.
  */
-size_t
+static size_t
 xmbstowcs(pwcs, s, nwcs)
 register wchar_t *pwcs;
 register const char *s;
@@ -111,9 +141,9 @@ size_t nwcs;
 	size_t n = nwcs;
 	register int c;
 
-	mbtowc(pwcs, NULL, MB_CUR_MAX);
+	mbtowc(pwcs, NULL, mb_cur_max);
 	while (*s && n) {
-		if ((c = mbtowc(pwcs, s, MB_CUR_MAX)) < 0) {
+		if ((c = mbtowc(pwcs, s, mb_cur_max)) < 0) {
 			s++;
 			*pwcs = L'?';
 		} else
@@ -123,7 +153,7 @@ size_t nwcs;
 	}
 	if (n)
 		*pwcs = L'\0';
-	mbtowc(pwcs, NULL, MB_CUR_MAX);
+	mbtowc(pwcs, NULL, mb_cur_max);
 	return nwcs - n;
 }
 #endif	/* HAVE_MBTOWC && HAVE_ISWPRINT */
@@ -140,8 +170,9 @@ size_t l;
 	size_t sz;
 #if defined (HAVE_MBTOWC) && defined (HAVE_ISWPRINT)
 	int i;
-	wchar_t w[LINESIZE], *p;
+	wchar_t *w, *p;
 	char *t;
+	size_t ret;
 
 #ifdef	__GLIBC_MINOR__
 #if __GLIBC__ <= 2 && __GLIBC_MINOR__ <= 1
@@ -150,32 +181,29 @@ size_t l;
 	 * Just a strange hack that will disappear one
 	 * day, but we need it for testing.
 	 */
-	if (strcasecmp(gettcharset(), "utf-8") == 0)
+	if (asccasecmp(gettcharset(), "utf-8") == 0)
 		return l;
 #endif
 #endif
 	if (*s == '\0' || l == (size_t) 0)
 		return 0;
-#ifdef	MB_CUR_MAX
-	if (MB_CUR_MAX > 1) {
-#endif
-		if (l >= LINESIZE)
-			return l;
+	if (mb_cur_max > 1) {
+		w = ac_alloc((l + 1) * sizeof *w);
 		t = s;
-		if ((sz = xmbstowcs(w, t, LINESIZE)) == (size_t)-1)
+		if ((sz = xmbstowcs(w, t, l + 1)) == (size_t)-1) {
+			ac_free(w);
 			return 0;
+		}
 		for (p = w, i = 0; *p && i < sz; p++, i++) {
 			if (!iswprint(*p) && *p != '\n' && *p != '\r'
 					&& *p != '\b' && *p != '\t')
 				*p = '?';
 		}
 		p = w;
-		return wcstombs(s, p, l + 1);
-#ifdef	MB_CUR_MAX
+		ret = wcstombs(s, p, l + 1);
+		ac_free(w);
+		return ret;
 	}
-#else	/* !MB_CUR_MAX */
-	/*NOTREACHED*/
-#endif	/* !MB_CUR_MAX */
 #endif	/* HAVE_MBTOWC && HAVE_ISWPRINT */
 	sz = l;
 	for (; l > 0 && *s; s++, l--) {
@@ -199,7 +227,7 @@ char *name;
 	char *addr, *p;
 	int in_quote = 0, in_domain = 0, err = 0, hadat = 0;
 
-	if (isfileaddr(name))
+	if (is_fileaddr(name))
 		return 0;
 	addr = skin(name);
 
@@ -213,7 +241,7 @@ char *name;
 			break;
 		} else if (in_domain == 2) {
 			if ((*p == ']' && p[1] != '\0') || *p == '\0'
-					|| *p == '\\' || blankchar(*p)) {
+					|| *p == '\\' || whitechar(*p & 0377)) {
 				err = *p & 0377;
 				break;
 			}
@@ -224,8 +252,9 @@ char *name;
 		} else if (*p == '@') {
 			if (hadat++) {
 				if (putmsg) {
-					fprintf(stderr,
-					"%s contains invalid @@ sequence\n",
+					fprintf(stderr, catgets(catd, CATSET,
+								142,
+					"%s contains invalid @@ sequence\n"),
 						addr);
 					putmsg = 0;
 				}
@@ -246,16 +275,17 @@ char *name;
 		hadat = 0;
 	}
 	if (err && putmsg) {
-		fprintf(stderr, "%s contains invalid character '", addr);
+		fprintf(stderr, catgets(catd, CATSET, 143,
+				"%s contains invalid character '"), addr);
 #ifdef	HAVE_SETLOCALE
 		if (isprint(err))
 #endif	/* HAVE_SETLOCALE */
-			fputc(err, stderr);
+			sputc(err, stderr);
 #ifdef	HAVE_SETLOCALE
 		else
 			fprintf(stderr, "\\%03o", err);
 #endif	/* HAVE_SETLOCALE */
-		fprintf(stderr, "'\n");
+		fprintf(stderr, catgets(catd, CATSET, 144, "'\n"));
 	}
 	return err;
 }
@@ -283,90 +313,22 @@ struct name *np;
 	return np;
 }
 
-static void
-out_of_memory()
-{
-	panic("Out of memory");
-}
-
-void *
-smalloc(s)
-size_t s;
-{
-	void *p;
-
-	if ((p = malloc(s)) == NULL) {
-		out_of_memory();
-	}
-	return p;
-}
-
-#ifndef	HAVE_STRCASECMP
-/* One of the things I really HATE on some SysVs is that they still
- * do not have these functions in their standard libc.
- * These implementations are quick-drawn and inefficient. Claim
- * your vendor for a better version, and one outside libucb!
- */
-int
-strcasecmp(s1, s2)
-const char *s1, *s2;
-{
-	int cmp, c1, c2;
-
-	do {
-		if (isupper(*s1))
-			c1 = tolower(*s1);
-		else
-			c1 = *s1 & 0377;
-		if (isupper(*s2))
-			c2 = tolower(*s2);
-		else
-			c2 = *s2 & 0377;
-		cmp = c1 - c2;
-		if (cmp != 0)
-			return cmp;
-	} while (*s1++ != '\0' && *s2++ != '\0');
-	return 0;
-}
-
-int
-strncasecmp(s1, s2, sz)
-const char *s1, *s2;
-size_t sz;
-{
-	int cmp, c1, c2;
-	size_t i = 1;
-
-	if (sz <= 0)
-		return 0;
-	do {
-		if (isupper(*s1))
-			c1 = tolower(*s1);
-		else
-			c1 = *s1 & 0377;
-		if (isupper(*s2))
-			c2 = tolower(*s2);
-		else
-			c2 = *s2 & 0377;
-		cmp = c1 - c2;
-		if (cmp != 0)
-			return cmp;
-	} while (i++ < sz && *s1++ != '\0' && *s2++ != '\0');
-	return 0;
-}
-#endif	/* silly SysV functions */
-
 static char defcharset[] = "iso-8859-1";
 
 /*
  * Get the character set dependant on the conversion.
  */
-char *
-getcharset(convert)
+static char *
+getcharset(isclean)
 {
 	char *charset;
 
-	if (convert == CONV_7BIT) {
+	if (isclean & MIME_HIGHBIT) {
+		charset = value("charset");
+		if (charset == NULL) {
+			charset = defcharset;
+		}
+	} else {
 		/*
 		 * This variable shall remain undocumented because
 		 * only experts should change it.
@@ -375,11 +337,6 @@ getcharset(convert)
 		if (charset == NULL) {
 			charset = us_ascii;
 		}
-	} else {
-		charset = value("charset");
-		if (charset == NULL) {
-			charset = defcharset;
-		}
 	}
 	return charset;
 }
@@ -398,27 +355,75 @@ gettcharset()
 	return t;
 }
 
+static int
+has_highbit(s)
+	register const char *s;
+{
+	if (s) {
+		do
+			if (*s & 0200)
+				return 1;
+		while (*s++ != '\0');
+	}
+	return 0;
+}
+
+char *
+need_hdrconv(hp, w)
+	struct header *hp;
+	enum gfield w;
+{
+	struct name *np;
+
+	if (w & GIDENT) {
+		if (has_highbit(myaddr()))
+			goto needs;
+		if (has_highbit(value("ORGANIZATION")))
+			goto needs;
+		if (has_highbit(value("replyto")))
+			goto needs;
+	}
+	if (w & GTO) {
+		for (np = hp->h_to; np; np = np->n_flink)
+			if (has_highbit(np->n_name))
+				goto needs;
+	}
+	if (w & GCC) {
+		for (np = hp->h_cc; np; np = np->n_flink)
+			if (has_highbit(np->n_name))
+				goto needs;
+	}
+	if (w & GBCC) {
+		for (np = hp->h_bcc; np; np = np->n_flink)
+			if (has_highbit(np->n_name))
+				goto needs;
+	}
+	if (w & GSUBJECT) {
+		if (has_highbit(hp->h_subject))
+			goto needs;
+	}
+	return NULL;
+needs:	return getcharset(MIME_HIGHBIT);
+}
+
 #ifdef	HAVE_ICONV
 /*
  * Convert a string, upper-casing the characters.
  */
-void
-strupcpy(dest, src)
+static void
+uppercopy(dest, src)
 char *dest;
 const char *src;
 {
 	do
-		if (islower(*src & 0377))
-			*dest++ = toupper(*src);
-		else
-			*dest++ = *src;
+		*dest++ = upperconv(*src & 0377);
 	while (*src++);
 }
 
 /*
  * Strip dashes.
  */
-void
+static void
 stripdash(p)
 char *p;
 {
@@ -453,13 +458,13 @@ const char *tocode, *fromcode;
 	/*
 	 * Remove the "iso-" prefixes for Solaris.
 	 */
-	if (strncasecmp(tocode, "iso-", 4) == 0)
+	if (ascncasecmp(tocode, "iso-", 4) == 0)
 		tocode += 4;
-	else if (strncasecmp(tocode, "iso", 3) == 0)
+	else if (ascncasecmp(tocode, "iso", 3) == 0)
 		tocode += 3;
-	if (strncasecmp(fromcode, "iso-", 4) == 0)
+	if (ascncasecmp(fromcode, "iso-", 4) == 0)
 		fromcode += 4;
-	else if (strncasecmp(fromcode, "iso", 3) == 0)
+	else if (ascncasecmp(fromcode, "iso", 3) == 0)
 		fromcode += 3;
 	if (*tocode == '\0' || *fromcode == '\0')
 		return (iconv_t) -1;
@@ -473,9 +478,9 @@ const char *tocode, *fromcode;
 	 * Solaris prefers upper-case charset names. Don't ask...
 	 */
 	t = (char *)salloc(strlen(tocode) + 1);
-	strupcpy(t, tocode);
+	uppercopy(t, tocode);
 	f = (char *)salloc(strlen(fromcode) + 1);
-	strupcpy(f, fromcode);
+	uppercopy(f, fromcode);
 	if (strcmp(t, f) == 0) {
 		errno = 0;
 		return (iconv_t)-1;
@@ -502,7 +507,7 @@ const char *tocode, *fromcode;
 /*
  * Fault-tolerant iconv() function.
  */
-size_t
+static size_t
 iconv_ft(cd, inb, inbleft, outb, outbleft)
 iconv_t cd;
 char **inb;
@@ -532,10 +537,27 @@ char **outb;
 }
 #endif	/* HAVE_ICONV */
 
+static int
+is_this_enc(const char *line, const char *encoding)
+{
+	int quoted = 0, c;
+
+	if (*line == '"')
+		quoted = 1, line++;
+	while (*line && *encoding)
+		if (c = *line++, lowerconv(c) != *encoding++)
+			return 0;
+	if (quoted && *line == '"')
+		return 1;
+	if (*line == '\0' || whitechar(*line & 0377))
+		return 1;
+	return 0;
+}
+
 /*
  * Get the mime encoding from a Content-Transfer-Encoding header line.
  */
-int
+enum mimeenc
 mime_getenc(h)
 char *h;
 {
@@ -544,17 +566,17 @@ char *h;
 	if ((p = strchr(h, ':')) == NULL)
 		return MIME_NONE;
 	p++;
-	while (*p && blankchar(*p))
+	while (*p && whitechar(*p & 0377))
 		p++;
-	if (strncasecmp(p, "7bit", 4) == 0)
+	if (is_this_enc(p, "7bit"))
 		return MIME_7B;
-	if (strncasecmp(p, "8bit", 4) == 0)
+	if (is_this_enc(p, "8bit"))
 		return MIME_8B;
-	if (strncasecmp(p, "base64", 6) == 0)
+	if (is_this_enc(p, "base64"))
 		return MIME_B64;
-	if (strncasecmp(p, "binary", 6) == 0)
+	if (is_this_enc(p, "binary"))
 		return MIME_BIN;
-	if (strncasecmp(p, "quoted-printable", 16) == 0)
+	if (is_this_enc(p, "quoted-printable"))
 		return MIME_QP;
 	return MIME_NONE;
 }
@@ -563,33 +585,34 @@ char *h;
  * Get the mime content from a Content-Type header line.
  */
 int
-mime_getcontent(h, ret)
+mime_getcontent(h)
 char *h;
-char **ret;
 {
-	char *p, *q, *s;
+	char *p, *q, *r, *s;
 
-	if ((s = strchr(h, ':')) == NULL) {
-		ret = NULL;
+	if ((s = strchr(h, ':')) == NULL)
 		return 1;
-	}
 	s++;
-	while (*s && blankchar(*s))
+	while (*s && whitechar(*s & 0377))
 		s++;
-	*ret = salloc(strlen(s) + 1);
-	p = s, q = *ret;
-	while (*p && !blankchar(*p) && *p != ';')
-		*q++ = tolower(*p++);
+	if (*s == '"')
+		s++;
+	r = salloc(strlen(s) + 1);
+	p = s, q = r;
+	while (*p && !whitechar(*p & 0377) && *p != ';') {
+		*q++ = lowerconv(*p & 0377);
+		p++;
+	}
 	*q = '\0';
-	if (strchr(*ret, '/') == NULL)	/* for compatibility with non-MIME */
+	if (strchr(r, '/') == NULL)	/* for compatibility with non-MIME */
 		return MIME_TEXT;
-	if (strncmp(*ret, "text/", 5) == 0)
+	if (strncmp(r, "text/", 5) == 0)
 		return MIME_TEXT;
-	if (strncmp(*ret, "message/rfc822", 14) == 0)
+	if (strncmp(r, "message/rfc822", 14) == 0)
 		return MIME_822;
-	if (strncmp(*ret, "message/", 8) == 0)
+	if (strncmp(r, "message/", 8) == 0)
 		return MIME_MESSAGE;
-	if (strncmp(*ret, "multipart/", 10) == 0)
+	if (strncmp(r, "multipart/", 10) == 0)
 		return MIME_MULTI;
 	return MIME_UNKNOWN;
 }
@@ -606,21 +629,21 @@ char *param, *h;
 	size_t sz;
 
 	sz = strlen(param);
-	if (!blankchar(*p)) {
+	if (!whitechar(*p & 0377)) {
 		c = '\0';
 		while (*p && (*p != ';' || c == '\\')) {
 			c = c == '\\' ? '\0' : *p;
-			*p++;
+			p++;
 		}
 		if (*p++ == '\0')
 			return NULL;
 	}
 	for (;;) {
-		while (blankchar(*p))
+		while (whitechar(*p & 0377))
 			p++;
-		if (strncasecmp(p, param, sz) == 0) {
+		if (ascncasecmp(p, param, sz) == 0) {
 			p += sz;
-			while (blankchar(*p))
+			while (whitechar(*p & 0377))
 				p++;
 			if (*p++ == '=')
 				break;
@@ -642,7 +665,7 @@ char *param, *h;
 		if (*p++ == '\0')
 			return NULL;
 	}
-	while (blankchar(*p))
+	while (whitechar(*p & 0377))
 		p++;
 	q = p;
 	c = '\0';
@@ -652,7 +675,7 @@ char *param, *h;
 			return NULL;
 	} else {
 		q = p;
-		while (*q && !blankchar(*q) && *q != ';')
+		while (*q && !whitechar(*q & 0377) && *q != ';')
 			q++;
 	}
 	sz = q - p;
@@ -682,16 +705,8 @@ char *h;
 	return q;
 }
 
-char *
-mime_getfilename(h)
-char *h;
-{
-	return mime_getparam("filename", h);
-}
-
 /*
  * Get a line like "text/html html" and look if x matches the extension.
- * The line must be terminated by a newline character.
  */
 static char *
 mime_tline(x, l)
@@ -700,30 +715,29 @@ char *x, *l;
 	char *type, *n;
 	int match = 0;
 
-	if ((*l & 0200) || isalpha(*l) == 0)
+	if ((*l & 0200) || alphachar(*l & 0377) == 0)
 		return NULL;
 	type = l;
-	while (spacechar(*l) == 0 && *l != '\0')
+	while (blankchar(*l & 0377) == 0 && *l != '\0')
 		l++;
 	if (*l == '\0')
 		return NULL;
 	*l++ = '\0';
-	while (spacechar(*l) != 0 && *l != '\0')
+	while (blankchar(*l & 0377) != 0 && *l != '\0')
 		l++;
 	if (*l == '\0')
 		return NULL;
 	while (*l != '\0') {
 		n = l;
-		while (blankchar(*l) == 0 && *l != '\0')
+		while (whitechar(*l & 0377) == 0 && *l != '\0')
 			l++;
-		if (*l == '\0')
-			return NULL;
-		*l++ = '\0';
+		if (*l != '\0')
+			*l++ = '\0';
 		if (strcmp(x, n) == 0) {
 			match = 1;
 			break;
 		}
-		while (spacechar(*l) != 0 && *l != '\0')
+		while (whitechar(*l & 0377) != 0 && *l != '\0')
 			l++;
 	}
 	if (match != 0) {
@@ -742,16 +756,19 @@ mime_type(ext, filename)
 char *ext, *filename;
 {
 	FILE *f;
-	char line[LINESIZE];
+	char *line = NULL;
+	size_t linesize = 0;
 	char *type = NULL;
 
 	if ((f = Fopen(filename, "r")) == NULL)
 		return NULL;
-	while (fgets(line, LINESIZE, f)) {
+	while (fgetline(&line, &linesize, NULL, NULL, f, 0)) {
 		if ((type = mime_tline(ext, line)) != NULL)
 			break;
 	}
 	Fclose(f);
+	if (line)
+		free(line);
 	return type;
 }
 
@@ -764,9 +781,8 @@ char *name;
 {
 	char *ext, *content;
 
-	if ((ext = strrchr(name, '.')) == NULL)
+	if ((ext = strrchr(name, '.')) == NULL || *++ext == '\0')
 		return NULL;
-	ext++;
 	if ((content = mime_type(ext, expand(mimetypes_user))) != NULL)
 		return content;
 	if ((content = mime_type(ext, mimetypes_world)) != NULL)
@@ -777,18 +793,18 @@ char *name;
 /*
  * Check file contents.
  */
-int
+static enum mimeclean
 mime_isclean(f)
 FILE *f;
 {
 	long initial_pos;
 	unsigned curlen = 1, maxlen = 0;
-	int isclean = MIME_7BITTEXT;
+	enum mimeclean isclean = 0;
 	int c;
 
 	initial_pos = ftell(f);
-	for (;;) {
-		c = getc(f);
+	do {
+		c = sgetc(f);
 		curlen++;
 		if (c == '\n' || c == EOF) {
 			/*
@@ -799,27 +815,84 @@ FILE *f;
 			if (curlen > maxlen)
 				maxlen = curlen;
 			curlen = 1;
-			if (c == '\n')
-				continue;
-			else
-				break;
-		}
-		if (c & 0200) {
-			isclean = MIME_INTERTEXT;
-			continue;
-		}
-		if ((c < 040 && (c != ' ' && c != '\t')) || c == 0177) {
-			isclean = MIME_BINARY;
+		} else if (c & 0200) {
+			isclean |= MIME_HIGHBIT;
+		} else if (c == '\0') {
+			isclean |= MIME_HASNUL;
 			break;
+		} else if ((c < 040 && (c != '\t' && c != '\f')) || c == 0177) {
+			isclean |= MIME_CTRLCHAR;
 		}
-	}
+	} while (c != EOF);
 	clearerr(f);
 	fseek(f, initial_pos, SEEK_SET);
 	if (maxlen > 950)
-		isclean = MIME_BINARY;
+		isclean |= MIME_LONGLINES;
 	return isclean;
 }
 
+/*
+ * Get the conversion that matches the encoding specified in the environment.
+ */
+static int
+gettextconversion()
+{
+	char *p;
+	int convert;
+
+	if ((p = value("encoding")) == NULL)
+		return CONV_NONE;
+	if (equal(p, "quoted-printable"))
+		convert = CONV_TOQP;
+	else if (equal(p, "8bit"))
+		convert = CONV_NONE;
+	else {
+		fprintf(stderr, catgets(catd, CATSET, 177,
+			"Warning: invalid encoding %s, using 8bit\n"), p);
+		convert = CONV_NONE;
+	}
+	return convert;
+}
+
+int
+get_mime_convert(fp, contenttype, charset, isclean)
+FILE *fp;
+char **contenttype;
+char **charset;
+enum mimeclean *isclean;
+{
+	int convert;
+
+	*isclean = mime_isclean(fp);
+	if (*isclean & MIME_HASNUL) {
+		convert = CONV_TOB64;
+		if (*contenttype == NULL ||
+				ascncasecmp(*contenttype, "text/", 5) == 0)
+			*contenttype = "application/octet-stream";
+		*charset = NULL;
+	} else if (*isclean & MIME_LONGLINES)
+		convert = CONV_TOQP;
+	else if (*isclean & MIME_HIGHBIT)
+		convert = gettextconversion();
+	else
+		convert = CONV_7BIT;
+	if (*contenttype == NULL) {
+		if (*isclean & MIME_CTRLCHAR) {
+			/*
+			 * RFC 2046 forbids control characters other than
+			 * ^I or ^L in text/plain bodies. However, some
+			 * obscure character sets actually contain these
+			 * characters, so the content type can be set.
+			 */
+			if ((*contenttype = value("contenttype-cntrl")) == NULL)
+				*contenttype = "application/octet-stream";
+		} else
+			*contenttype = "text/plain";
+		*charset = getcharset(*isclean);
+	}
+	return convert;
+}
+
 /*
  * Print an error because of an invalid character sequence.
  */
@@ -895,14 +968,14 @@ int (*mustquote)(int);
 	upper = in->s + in->l;
 	for (p = in->s, l = 0; p < upper; p++) {
 		if (mustquote(*p) 
-				|| (*(p + 1) == '\n' && spacechar(*p))) {
+				|| (*(p + 1) == '\n' && blankchar(*p & 0377))) {
 			if (l >= 69) {
 				sz += 2;
 				fwrite("=\n", sizeof (char), 2, fo);
 				l = 0;
 			}
 			sz += 2;
-			fputc('=', fo);
+			sputc('=', fo);
 			h = ctohex(*p, hex);
 			fwrite(h, sizeof *h, 2, fo);
 			l += 3;
@@ -914,7 +987,7 @@ int (*mustquote)(int);
 				fwrite("=\n", sizeof (char), 2, fo);
 				l = 0;
 			}
-			fputc(*p, fo);
+			sputc(*p, fo);
 			l++;
 		}
 	}
@@ -937,7 +1010,8 @@ int (*mustquote)(int);
 	out->l = in->l;
 	upper = in->s + in->l;
 	for (p = in->s; p < upper; p++) {
-		if (mustquote(*p) || (*(p + 1) == '\n' && spacechar(*p))) {
+		if (mustquote(*p) ||
+				(*(p + 1) == '\n' && blankchar(*p & 0377))) {
 			out->l += 2;
 			*q++ = '=';
 			ctohex(*p, q);
@@ -967,7 +1041,7 @@ struct str *in, *out;
 			do {
 				p++;
 				out->l--;
-			} while (spacechar(*p) && p < upper);
+			} while (blankchar(*p & 0377) && p < upper);
 			if (p == upper)
 				break;
 			if (*p == '\n') {
@@ -990,12 +1064,18 @@ struct str *in, *out;
 	return;
 }
 
+#define	mime_fromhdr_inc(inc) { \
+		size_t diff = q - out->s; \
+		out->s = srealloc(out->s, (maxstor += inc) + 1); \
+		q = &(out->s)[diff]; \
+	}
 /*
  * Convert header fields from RFC 1522 format
  */
 void
 mime_fromhdr(in, out, flags)
 struct str *in, *out;
+enum tdflags flags;
 {
 	char *p, *q, *op, *upper, *cs, *cbeg, *tcs;
 	struct str cin, cout;
@@ -1006,8 +1086,8 @@ struct str *in, *out;
 #endif
 
 	tcs = gettcharset();
-	maxstor = 4 * in->l;
-	out->s = (char *)smalloc(maxstor + 1);
+	maxstor = in->l;
+	out->s = smalloc(maxstor + 1);
 	out->l = 0;
 	upper = in->s + in->l;
 	for (p = in->s, q = out->s; p < upper; p++) {
@@ -1059,25 +1139,30 @@ struct str *in, *out;
 				case CONV_FROMQP:
 					mime_fromqp(&cin, &cout, 1);
 					break;
-				default: abort();
 			}
 #ifdef	HAVE_ICONV
 			if ((flags & TD_ICONV) && fhicd != (iconv_t)-1) {
 				char *iptr, *mptr, *nptr, *uptr;
 				size_t inleft, outleft;
 
-				inleft = cout.l;
+			again:	inleft = cout.l;
 				outleft = maxstor - out->l;
 				mptr = nptr = q;
 				uptr = nptr + outleft;
 				iptr = cout.s;
-				iconv_ft(fhicd, (const char **)&iptr,
-							&inleft, &nptr,
-							&outleft);
+				if (iconv_ft(fhicd, &iptr, &inleft,
+						&nptr, &outleft) != 0 &&
+						errno == E2BIG) {
+					mime_fromhdr_inc(inleft);
+					goto again;
+				}
 				out->l += uptr - mptr - outleft;
 				q += uptr - mptr - outleft;
 			} else {
 #endif
+				while (cout.l > maxstor - out->l)
+					mime_fromhdr_inc(cout.l -
+							(maxstor - out->l));
 				memcpy(q, cout.s, cout.l);
 				q += cout.l;
 				out->l += cout.l;
@@ -1088,6 +1173,8 @@ struct str *in, *out;
 		} else {
 notmime:
 			p = op;
+			while (out->l >= maxstor)
+				mime_fromhdr_inc(16);
 			*q++ = *p;
 			out->l++;
 		}
@@ -1118,7 +1205,7 @@ FILE *fo;
 		maxcol = 65 /* there is the header field's name, too */;
 
 	upper = in->s + in->l;
-	charset = getcharset(CONV_TOQP);
+	charset = getcharset(MIME_HIGHBIT);
 	charsetlen = strlen(charset);
 	for (wbeg = in->s, mustquote = 0; wbeg < upper; wbeg++)
 		if (mustquote_hdr(*wbeg))
@@ -1163,14 +1250,15 @@ FILE *fo;
 		 * Print the field word-wise in quoted-printable.
 		 */
 		for (wbeg = in->s; wbeg < upper; wbeg = wend) {
-			while (wbeg < upper && blankchar(*wbeg)) {
-				fputc(*wbeg++, fo);
+			while (wbeg < upper && whitechar(*wbeg & 0377)) {
+				sputc(*wbeg++, fo);
 				sz++, col++;
 			}
 			if (wbeg == upper)
 				break;
 			mustquote = 0;
-			for (wend = wbeg; wend < upper && !blankchar(*wend);
+			for (wend = wbeg;
+				wend < upper && !whitechar(*wend & 0377);
 					wend++) {
 				if (mustquote_hdr(*wend))
 					mustquote++;
@@ -1230,88 +1318,101 @@ FILE *f;
 {
 	char *p, *q, *ip, *op;
 	struct str cin;
-	size_t sz = 0, isz, osz;
-	char cbuf[LINESIZE];
-#ifdef	HAVE_ICONV
-	iconv_t fhicd = (iconv_t)-1;
-#endif
+	size_t sz = 0, isz, osz, cbufsz;
+	char *cbuf;
 
+	cbuf = ac_alloc(cbufsz = 1);
 	*(in->s + in->l) = '\0';
-	if (strpbrk(in->s, "(<") == NULL)
+	if (strpbrk(in->s, "(<") == NULL) {
+		ac_free(cbuf);
 		return fwrite(in->s, sizeof *in->s, in->l, f);
+	}
 	if ((p = strchr(in->s, '<')) != NULL) {
 		q = strchr(p, '>');
-		if (q++ == NULL)
+		if (q++ == NULL) {
+			ac_free(cbuf);
 			return 0;
+		}
 #ifdef	HAVE_ICONV
-		if (fhicd == (iconv_t)-1) {
+		if (iconvd == (iconv_t)-1) {
 #endif
 			cin.s = in->s;
 			cin.l = p - in->s;
 #ifdef	HAVE_ICONV
 		} else {
-			ip = in->s;
+		again1:	ip = in->s;
 			isz = p - in->s;
 			op = cbuf;
-			osz = sizeof cbuf;
-			if (iconv(fhicd, &ip, &isz, &op, &osz) == (size_t)-1) {
-				iconv_close(fhicd);
-				return 0;
+			osz = cbufsz;
+			if (iconv(iconvd, &ip, &isz, &op, &osz) != 0) {
+				if (errno != E2BIG) {
+					ac_free(cbuf);
+					return 0;
+				}
+				cbuf = ac_alloc(cbufsz += isz);
+				goto again1;
 			}
 			cin.s = cbuf;
-			cin.l = sizeof cbuf - osz;
+			cin.l = cbufsz - osz;
 		}
 #endif
 		sz = mime_write_tohdr(&cin, f);
 		sz += fwrite(p, sizeof *p, q - p, f);
 #ifdef	HAVE_ICONV
-		if (fhicd == (iconv_t)-1) {
+		if (iconvd == (iconv_t)-1) {
 #endif
 			cin.s = q;
 			cin.l = in->s + in->l - q;
 #ifdef	HAVE_ICONV
 		} else {
-			ip = q;
+		again2:	ip = q;
 			isz = in->s + in->l - q;
 			op = cbuf;
-			osz = sizeof cbuf;
-			if (iconv(fhicd, &ip, &isz, &op, &osz) == (size_t)-1)
-				return 0;
+			osz = cbufsz;
+			if (iconv(iconvd, &ip, &isz, &op, &osz) != 0) {
+				if (errno != E2BIG) {
+					ac_free(cbuf);
+					return 0;
+				}
+				cbuf = ac_alloc(cbufsz += isz);
+				goto again2;
+			}
 			cin.s = cbuf;
-			cin.l = sizeof cbuf - osz;
+			cin.l = cbufsz - osz;
 		}
 #endif
 		sz += mime_write_tohdr(&cin, f);
 	} else if ((p = strchr(in->s, '(')) != NULL) {
+		p++;
 		q = in->s;
 		do {
 			sz += fwrite(q, sizeof *q, p - q, f);
 			q = strchr(p, ')');
 			if (q++ == NULL) {
-#ifdef	HAVE_ICONV
-				if (fhicd != (iconv_t)-1)
-					iconv_close(fhicd);
-#endif
+				ac_free(cbuf);
 				return 0;
 			}
 #ifdef	HAVE_ICONV
-			if (fhicd == (iconv_t)-1) {
+			if (iconvd == (iconv_t)-1) {
 #endif
 				cin.s = p;
 				cin.l = q - p;
 #ifdef	HAVE_ICONV
 			} else {
-				ip = p;
+			again3:	ip = p;
 				isz = q - p;
 				op = cbuf;
-				osz = sizeof cbuf;
-				if (iconv(fhicd, &ip, &isz,
-					&op, &osz) == (size_t)-1) {
-					iconv_close(fhicd);
-					return 0;
+				osz = cbufsz;
+				if (iconv(iconvd, &ip, &isz, &op, &osz) != 0) {
+					if (errno != E2BIG) {
+						ac_free(cbuf);
+						return 0;
+					}
+					cbuf = ac_alloc(cbufsz += isz);
+					goto again3;
 				}
 				cin.s = cbuf;
-				cin.l = sizeof cbuf - osz;
+				cin.l = cbufsz - osz;
 			}
 #endif
 			sz += mime_write_tohdr(&cin, f);
@@ -1320,10 +1421,7 @@ FILE *f;
 		if (*q != '\0')
 			sz += fwrite(q, sizeof *q, in->s + in->l - q, f);
 	}
-#ifdef	HAVE_ICONV
-	if (fhicd != (iconv_t)-1)
-		iconv_close(fhicd);
-#endif
+	ac_free(cbuf);
 	return sz;
 }
 
@@ -1359,7 +1457,7 @@ char *prefix;
 	}
 	lastf = f;
 	for (rsz = size * nmemb; rsz; rsz--, p++, wsz++) {
-		fputc(*p, f);
+		sputc(*p, f);
 		if (*p != '\n' || rsz == 1) {
 			continue;
 		}
@@ -1383,46 +1481,51 @@ void *ptr;
 char *prefix;
 size_t size, nmemb, prefixlen;
 FILE *f;
+enum tdflags flags;
 {
 	char *upper;
 	size_t sz, csize;
 #ifdef	HAVE_ICONV
 	char *iptr, *nptr;
 	size_t inleft, outleft;
-	char mptr[LINESIZE * 4];
-#else
-	char mptr[LINESIZE];
 #endif
+	char *mptr;
+	size_t mptrsz;
 
-	csize = size * nmemb / sizeof (char);
-	if (flags == 0 || csize >= sizeof mptr)
-		return prefixwrite(ptr, size, nmemb, f, prefix, prefixlen);
+	csize = size * nmemb;
+	mptrsz = csize;
+	mptr = ac_alloc(mptrsz + 1);
 #ifdef	HAVE_ICONV
 	if ((flags & TD_ICONV) && iconvd != (iconv_t)-1) {
-		inleft = csize;
-		outleft = sizeof mptr;
+	again:	inleft = csize;
+		outleft = mptrsz;
 		nptr = mptr;
 		iptr = ptr;
-		iconv_ft(iconvd, (const char **)&iptr, &inleft,
-				&nptr, &outleft);
-		nmemb = sizeof mptr - outleft;
+		if (iconv_ft(iconvd, &iptr, &inleft, &nptr, &outleft) != 0 &&
+				errno == E2BIG) {
+			ac_free(mptr);
+			mptrsz += inleft;
+			mptr = ac_alloc(mptrsz + 1);
+			goto again;
+		}
+		nmemb = mptrsz - outleft;
 		size = sizeof (char);
 		ptr = mptr;
-		csize = size * nmemb / sizeof (char);
+		csize = size * nmemb;
 	} else
 #endif
 	{
-		memcpy(mptr, ptr, nmemb / size);
+		memcpy(mptr, ptr, csize);
 	}
-	upper = (char *)mptr + csize;
+	upper = mptr + csize;
 	*upper = '\0';
 	if (flags & TD_ISPR) {
-		if ((csize = makeprint(mptr, upper - (char *)mptr)) >
-				upper - (char *)mptr)
-			csize = upper - (char *)mptr;
-	} else
-		csize = upper - (char *)mptr;
+		if ((csize = makeprint(mptr, upper - mptr)) >
+				upper - mptr)
+			csize = upper - mptr;
+	}
 	sz = prefixwrite(mptr, sizeof *mptr, csize, f, prefix, prefixlen);
+	ac_free(mptr);
 	return sz;
 }
 
@@ -1435,23 +1538,26 @@ void *ptr;
 size_t size, nmemb, prefixlen;
 char *prefix;
 FILE *f;
+enum conversion convert;
+enum tdflags dflags;
 {
 	struct str in, out;
 	size_t sz, csize;
 	int is_text = 0;
 #ifdef	HAVE_ICONV
-	char mptr[LINESIZE * 4];
+	char mptr[LINESIZE * 6];
 	char *iptr, *nptr;
 	size_t inleft, outleft;
 #endif
 
+	if (size == 0 || nmemb == 0)
+		return 0;
 	csize = size * nmemb / sizeof (char);
 #ifdef	HAVE_ICONV
 	if (csize < sizeof mptr && (dflags & TD_ICONV)
 			&& iconvd != (iconv_t)-1
 			&& (convert == CONV_TOQP || convert == CONV_TOB64
 				|| convert == CONV_TOHDR)) {
-
 		inleft = csize;
 		outleft = sizeof mptr;
 		nptr = mptr;
