@@ -1,3 +1,8 @@
+/*
+ * Nail - a mail user agent derived from Berkeley Mail.
+ *
+ * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ */
 /*
  * Copyright (c) 1980, 1993
  *	The Regents of the University of California.  All rights reserved.
@@ -33,7 +38,7 @@
 
 #ifndef lint
 #ifdef	DOSCCS
-static char sccsid[] = "@(#)quit.c	1.14 (gritter) 5/22/02";
+static char sccsid[] = "@(#)quit.c	2.6 (gritter) 10/27/02";
 #endif
 #endif /* not lint */
 
@@ -50,6 +55,7 @@ static char sccsid[] = "@(#)quit.c	1.14 (gritter) 5/22/02";
  *
  * Termination processing.
  */
+static void	edstop __P((void));
 
 /*
  * The "quit" command.
@@ -100,14 +106,14 @@ writeback(res, obuf)
 #endif
 #ifndef APPEND
 	if (res != (FILE*)NULL)
-		while ((c = getc(res)) != EOF)
-			(void) putc(c, obuf);
+		while ((c = sgetc(res)) != EOF)
+			(void) sputc(c, obuf);
 #endif
 	for (mp = &message[0]; mp < &message[msgcount]; mp++)
 		if ((mp->m_flag&MPRESERVE)||(mp->m_flag&MTOUCH)==0) {
 			p++;
 			if (send_message(mp, obuf, (struct ignoretab *)0,
-						NULL, CONV_NONE) < 0) {
+						NULL, CONV_NONE, NULL) < 0) {
 				perror(mailname);
 #ifndef	F_SETLKW
 				Fclose(obuf);
@@ -119,8 +125,8 @@ writeback(res, obuf)
 		}
 #ifdef APPEND
 	if (res != (FILE*)NULL)
-		while ((c = getc(res)) != EOF)
-			(void) putc(c, obuf);
+		while ((c = sgetc(res)) != EOF)
+			(void) sputc(c, obuf);
 #endif
 	fflush(obuf);
 	trunc(obuf);
@@ -142,9 +148,11 @@ writeback(res, obuf)
 #endif
 	alter(mailname);
 	if (p == 1)
-		printf("Held 1 message in %s\n", mailname);
+		printf(catgets(catd, CATSET, 155,
+				"Held 1 message in %s\n"), mailname);
 	else
-		printf("Held %d messages in %s\n", p, mailname);
+		printf(catgets(catd, CATSET, 156,
+				"Held %d messages in %s\n"), p, mailname);
 	return(0);
 }
 
@@ -156,20 +164,28 @@ writeback(res, obuf)
 void
 quit()
 {
-	int mcount, p, modify, autohold, anystat, holdbit, nohold;
-	FILE *ibuf = (FILE*)NULL, *obuf, *fbuf, *rbuf, *readstat = (FILE*)NULL, *abuf;
+	int p, modify, anystat;
+	FILE *fbuf, *rbuf, *readstat = (FILE*)NULL, *abuf;
 	struct message *mp;
 	int c;
-	char *tempQuit, *tempResid;
+	char *tempResid;
 	struct stat minfo;
-	char *mbox;
 
 	/*
 	 * If we are read only, we can't do anything,
 	 * so just return quickly.
 	 */
-	if (readonly)
+	if (mb.mb_perm == 0)
 		return;
+	switch (mb.mb_type) {
+	case MB_FILE:
+		break;
+	case MB_POP3:
+		pop3_quit();
+		return;
+	case MB_VOID:
+		return;
+	}
 	/*
 	 * If editing (not reading system mail box), then do the work
 	 * in edstop()
@@ -205,7 +221,7 @@ quit()
 	if (fcntl_lock(fileno(fbuf), F_WRLCK) == -1) {
 #endif
 nolock:
-		perror("Unable to lock mailbox");
+		perror(catgets(catd, CATSET, 157, "Unable to lock mailbox"));
 		Fclose(fbuf);
 		return;
 	}
@@ -213,21 +229,21 @@ nolock:
 		goto nolock;
 	rbuf = (FILE *) NULL;
 	if (fstat(fileno(fbuf), &minfo) >= 0 && minfo.st_size > mailsize) {
-		printf("New mail has arrived.\n");
-		rbuf = Ftemp(&tempResid, "Rq", "w", 0600);
+		printf(catgets(catd, CATSET, 158, "New mail has arrived.\n"));
+		rbuf = Ftemp(&tempResid, "Rq", "w", 0600, 1);
 		if (rbuf == (FILE*)NULL || fbuf == (FILE*)NULL)
 			goto newmail;
 #ifdef APPEND
 		fseek(fbuf, (long)mailsize, SEEK_SET);
-		while ((c = getc(fbuf)) != EOF)
-			(void) putc(c, rbuf);
+		while ((c = sgetc(fbuf)) != EOF)
+			(void) sputc(c, rbuf);
 #else
 		p = minfo.st_size - mailsize;
 		while (p-- > 0) {
-			c = getc(fbuf);
+			c = sgetc(fbuf);
 			if (c == EOF)
 				goto newmail;
-			(void) putc(c, rbuf);
+			(void) sputc(c, rbuf);
 		}
 #endif
 		Fclose(rbuf);
@@ -237,28 +253,7 @@ nolock:
 		Ftfree(&tempResid);
 	}
 
-	/*
-	 * Adjust the message flags in each message.
-	 */
-
-	anystat = 0;
-	autohold = value("hold") != NULL;
-	holdbit = autohold ? MPRESERVE : MBOX;
-	nohold = MBOX|MSAVED|MDELETED|MPRESERVE;
-	if (value("keepsave") != NULL)
-		nohold &= ~MSAVED;
-	for (mp = &message[0]; mp < &message[msgcount]; mp++) {
-		if (mp->m_flag & MNEW) {
-			mp->m_flag &= ~MNEW;
-			mp->m_flag |= MSTATUS;
-		}
-		if (mp->m_flag & MSTATUS)
-			anystat++;
-		if ((mp->m_flag & MTOUCH) == 0)
-			mp->m_flag |= MPRESERVE;
-		if ((mp->m_flag & nohold) == 0)
-			mp->m_flag |= holdbit;
-	}
+	anystat = holdbits();
 	modify = 0;
 	if (Tflag != NULL) {
 		if ((readstat = Fopen(Tflag, "w")) == (FILE*)NULL)
@@ -271,18 +266,23 @@ nolock:
 			p++;
 		if (mp->m_flag & MODIFY)
 			modify++;
-		if (Tflag != NULL && (mp->m_flag & (MREAD|MDELETED)) != 0) {
+		if (readstat != NULL && (mp->m_flag & (MREAD|MDELETED)) != 0) {
 			char *id;
 
-			if ((id = hfield("article-id", mp)) != NULL)
+			if ((id = hfield("message-id", mp)) != NULL ||
+					(id = hfield("article-id", mp)) != NULL)
 				fprintf(readstat, "%s\n", id);
 		}
 	}
-	if (Tflag != NULL)
+	if (readstat != NULL)
 		Fclose(readstat);
 	if (p == msgcount && !modify && !anystat) {
-		printf("Held %d message%s in %s\n",
-			p, p == 1 ? "" : "s", mailname);
+		if (p == 1)
+			printf(catgets(catd, CATSET, 155,
+				"Held 1 message in %s\n"), mailname);
+		else if (p > 1)
+			printf(catgets(catd, CATSET, 156,
+				"Held %d messages in %s\n"), p, mailname);
 		Fclose(fbuf);
 		dot_unlock(mailname);
 		return;
@@ -301,110 +301,11 @@ nolock:
 		goto cream;
 	}
 
-	/*
-	 * Create another temporary file and copy user's mbox file
-	 * darin.  If there is no mbox, copy nothing.
-	 * If he has specified "append" don't copy his mailbox,
-	 * just copy saveable entries at the end.
-	 */
-
-	mbox = expand("&");
-	mcount = c;
-	if (value("append") == NULL) {
-		if ((obuf = Ftemp(&tempQuit, "Rm", "w", 0600)) == (FILE*)NULL) {
-			perror("temporary mail quit file");
-			Fclose(fbuf);
-			dot_unlock(mailname);
-			return;
-		}
-		if ((ibuf = Fopen(tempQuit, "r")) == (FILE*)NULL) {
-			perror(tempQuit);
-			rm(tempQuit);
-			Ftfree(&tempQuit);
-			Fclose(obuf);
-			Fclose(fbuf);
-			dot_unlock(mailname);
-			return;
-		}
-		rm(tempQuit);
-		Ftfree(&tempQuit);
-		if ((abuf = Fopen(mbox, "r")) != (FILE*)NULL) {
-			while ((c = getc(abuf)) != EOF)
-				(void) putc(c, obuf);
-			Fclose(abuf);
-		}
-		if (ferror(obuf)) {
-			perror("temporary mail quit file");
-			Fclose(ibuf);
-			Fclose(obuf);
-			Fclose(fbuf);
-			dot_unlock(mailname);
-			return;
-		}
-		Fclose(obuf);
-		close(creat(mbox, 0600));
-		if ((obuf = Fopen(mbox, "r+")) == (FILE*)NULL) {
-			perror(mbox);
-			Fclose(ibuf);
-			Fclose(fbuf);
-			dot_unlock(mailname);
-			return;
-		}
-	}
-	else {
-		if ((obuf = Fopen(mbox, "a")) == (FILE*)NULL) {
-			perror(mbox);
-			Fclose(fbuf);
-			dot_unlock(mailname);
-			return;
-		}
-		fchmod(fileno(obuf), 0600);
-	}
-	for (mp = &message[0]; mp < &message[msgcount]; mp++)
-		if (mp->m_flag & MBOX)
-			if (send_message(mp, obuf, saveignore,
-						NULL, CONV_NONE) < 0) {
-				perror(mbox);
-				if (ibuf)
-					Fclose(ibuf);
-				Fclose(obuf);
-				Fclose(fbuf);
-				dot_unlock(mailname);
-				return;
-			}
-
-	/*
-	 * Copy the user's old mbox contents back
-	 * to the end of the stuff we just saved.
-	 * If we are appending, this is unnecessary.
-	 */
-
-	if (value("append") == NULL) {
-		rewind(ibuf);
-		c = getc(ibuf);
-		while (c != EOF) {
-			(void) putc(c, obuf);
-			if (ferror(obuf))
-				break;
-			c = getc(ibuf);
-		}
-		Fclose(ibuf);
-		fflush(obuf);
-	}
-	trunc(obuf);
-	if (ferror(obuf)) {
-		perror(mbox);
-		Fclose(obuf);
+	if (makembox() == STOP) {
 		Fclose(fbuf);
 		dot_unlock(mailname);
 		return;
 	}
-	Fclose(obuf);
-	if (mcount == 1)
-		printf("Saved 1 message in mbox\n");
-	else
-		printf("Saved %d messages in mbox\n", mcount);
-
 	/*
 	 * Now we are ready to copy back preserved files to
 	 * the system mailbox, if any were requested.
@@ -436,8 +337,8 @@ cream:
 		abuf = fbuf;
 		fseek(abuf, 0L, SEEK_SET);
 #endif
-		while ((c = getc(rbuf)) != EOF)
-			(void) putc(c, abuf);
+		while ((c = sgetc(rbuf)) != EOF)
+			(void) sputc(c, abuf);
 		Fclose(rbuf);
 		trunc(abuf);
 #ifndef	F_SETLKW
@@ -454,26 +355,161 @@ cream:
 	return;
 
 newmail:
-	printf("Thou hast new mail.\n");
+	printf(catgets(catd, CATSET, 166, "Thou hast new mail.\n"));
 	if (fbuf != (FILE*)NULL) {
 		Fclose(fbuf);
 		dot_unlock(mailname);
 	}
 }
 
+/*
+ * Adjust the message flags in each message.
+ */
+int
+holdbits()
+{
+	struct message *mp;
+	int anystat, autohold, holdbit, nohold;
+
+	anystat = 0;
+	autohold = value("hold") != NULL;
+	holdbit = autohold ? MPRESERVE : MBOX;
+	nohold = MBOX|MSAVED|MDELETED|MPRESERVE;
+	if (value("keepsave") != NULL)
+		nohold &= ~MSAVED;
+	for (mp = &message[0]; mp < &message[msgcount]; mp++) {
+		if (mp->m_flag & MNEW) {
+			mp->m_flag &= ~MNEW;
+			mp->m_flag |= MSTATUS;
+		}
+		if (mp->m_flag & MSTATUS)
+			anystat++;
+		if ((mp->m_flag & MTOUCH) == 0)
+			mp->m_flag |= MPRESERVE;
+		if ((mp->m_flag & nohold) == 0)
+			mp->m_flag |= holdbit;
+	}
+	return anystat;
+}
+
+/*
+ * Create another temporary file and copy user's mbox file
+ * darin.  If there is no mbox, copy nothing.
+ * If he has specified "append" don't copy his mailbox,
+ * just copy saveable entries at the end.
+ */
+
+enum okay
+makembox()
+{
+	struct message *mp;
+	char *mbox, *tempQuit;
+	int mcount, c;
+	FILE *ibuf = NULL, *obuf, *abuf;
+
+	mbox = expand("&");
+	mcount = 0;
+	if (value("append") == NULL) {
+		if ((obuf = Ftemp(&tempQuit, "Rm", "w", 0600, 1)) == NULL) {
+			perror(catgets(catd, CATSET, 162,
+					"temporary mail quit file"));
+			return STOP;
+		}
+		if ((ibuf = Fopen(tempQuit, "r")) == NULL) {
+			perror(tempQuit);
+			rm(tempQuit);
+			Ftfree(&tempQuit);
+			Fclose(obuf);
+			return STOP;
+		}
+		rm(tempQuit);
+		Ftfree(&tempQuit);
+		if ((abuf = Fopen(mbox, "r")) != (FILE*)NULL) {
+			while ((c = sgetc(abuf)) != EOF)
+				(void) sputc(c, obuf);
+			Fclose(abuf);
+		}
+		if (ferror(obuf)) {
+			perror(catgets(catd, CATSET, 163,
+					"temporary mail quit file"));
+			Fclose(ibuf);
+			Fclose(obuf);
+			return STOP;
+		}
+		Fclose(obuf);
+		close(creat(mbox, 0600));
+		if ((obuf = Fopen(mbox, "r+")) == (FILE*)NULL) {
+			perror(mbox);
+			Fclose(ibuf);
+			return STOP;
+		}
+	}
+	else {
+		if ((obuf = Fopen(mbox, "a")) == (FILE*)NULL) {
+			perror(mbox);
+			return STOP;
+		}
+		fchmod(fileno(obuf), 0600);
+	}
+	for (mp = &message[0]; mp < &message[msgcount]; mp++)
+		if (mp->m_flag & MBOX) {
+			mcount++;
+			if (send_message(mp, obuf, saveignore,
+						NULL, CONV_NONE, NULL) < 0) {
+				perror(mbox);
+				if (ibuf)
+					Fclose(ibuf);
+				Fclose(obuf);
+				return STOP;
+			}
+		}
+
+	/*
+	 * Copy the user's old mbox contents back
+	 * to the end of the stuff we just saved.
+	 * If we are appending, this is unnecessary.
+	 */
+
+	if (value("append") == NULL) {
+		rewind(ibuf);
+		c = sgetc(ibuf);
+		while (c != EOF) {
+			(void) sputc(c, obuf);
+			if (ferror(obuf))
+				break;
+			c = sgetc(ibuf);
+		}
+		Fclose(ibuf);
+		fflush(obuf);
+	}
+	trunc(obuf);
+	if (ferror(obuf)) {
+		perror(mbox);
+		Fclose(obuf);
+		return STOP;
+	}
+	Fclose(obuf);
+	if (mcount == 1)
+		printf(catgets(catd, CATSET, 164, "Saved 1 message in mbox\n"));
+	else
+		printf(catgets(catd, CATSET, 165,
+				"Saved %d messages in mbox\n"), mcount);
+	return OKAY;
+}
+
 /*
  * Terminate an editing session by attempting to write out the user's
  * file from the temporary.  Save any new stuff appended to the file.
  */
-void
+static void
 edstop()
 {
 	int gotcha, c;
 	struct message *mp;
-	FILE *obuf, *ibuf, *readstat = (FILE*)NULL;
+	FILE *obuf, *ibuf = NULL, *readstat = (FILE*)NULL;
 	struct stat statb;
 
-	if (readonly)
+	if (mb.mb_perm == 0)
 		return;
 	holdsigs();
 	if (Tflag != NULL) {
@@ -487,14 +523,15 @@ edstop()
 		}
 		if (mp->m_flag & (MODIFY|MDELETED|MSTATUS))
 			gotcha++;
-		if (Tflag != NULL && (mp->m_flag & (MREAD|MDELETED)) != 0) {
+		if (readstat != NULL && (mp->m_flag & (MREAD|MDELETED)) != 0) {
 			char *id;
 
-			if ((id = hfield("article-id", mp)) != NULL)
+			if ((id = hfield("message-id", mp)) != NULL ||
+					(id = hfield("article-id", mp)) != NULL)
 				fprintf(readstat, "%s\n", id);
 		}
 	}
-	if (Tflag != NULL)
+	if (readstat != NULL)
 		Fclose(readstat);
 	if (!gotcha || Tflag != NULL)
 		goto done;
@@ -502,9 +539,8 @@ edstop()
 	if (stat(mailname, &statb) >= 0 && statb.st_size > mailsize) {
 		char *tempname;
 
-		if ((obuf = Ftemp(&tempname, "mbox.", "w", 0600))
-				== (FILE *)NULL) {
-			perror("tmpfile");
+		if ((obuf = Ftemp(&tempname, "mbox.", "w", 0600, 1)) == NULL) {
+			perror(catgets(catd, CATSET, 167, "tmpfile"));
 			relsesigs();
 			reset(0);
 		}
@@ -517,8 +553,8 @@ edstop()
 			reset(0);
 		}
 		fseek(ibuf, (long)mailsize, SEEK_SET);
-		while ((c = getc(ibuf)) != EOF)
-			(void) putc(c, obuf);
+		while ((c = sgetc(ibuf)) != EOF)
+			(void) sputc(c, obuf);
 		Fclose(ibuf);
 		Fclose(obuf);
 		if ((ibuf = Fopen(tempname, "r")) == (FILE*)NULL) {
@@ -531,7 +567,7 @@ edstop()
 		rm(tempname);
 		Ftfree(&tempname);
 	}
-	printf("\"%s\" ", mailname);
+	printf(catgets(catd, CATSET, 168, "\"%s\" "), mailname);
 	fflush(stdout);
 	if ((obuf = Fopen(mailname, "r+")) == (FILE*)NULL) {
 		perror(mailname);
@@ -545,7 +581,7 @@ edstop()
 			continue;
 		c++;
 		if (send_message(mp, obuf, (struct ignoretab *) NULL,
-					NULL, CONV_NONE) < 0) {
+					NULL, CONV_NONE, NULL) < 0) {
 			perror(mailname);
 			relsesigs();
 			reset(0);
@@ -553,8 +589,8 @@ edstop()
 	}
 	gotcha = (c == 0 && ibuf == (FILE*)NULL);
 	if (ibuf != (FILE*)NULL) {
-		while ((c = getc(ibuf)) != EOF)
-			(void) putc(c, obuf);
+		while ((c = sgetc(ibuf)) != EOF)
+			(void) sputc(c, obuf);
 		Fclose(ibuf);
 	}
 	fflush(obuf);
@@ -566,9 +602,13 @@ edstop()
 	Fclose(obuf);
 	if (gotcha && value("emptybox") == NULL) {
 		rm(mailname);
-		printf("removed\n");
+		printf(value("bsdcompat") ?
+				catgets(catd, CATSET, 169, "removed\n") :
+				catgets(catd, CATSET, 211, "removed.\n"));
 	} else
-		printf("complete\n");
+		printf(value("bsdcompat") ?
+				catgets(catd, CATSET, 170, "complete\n") :
+				catgets(catd, CATSET, 212, "updated.\n"));
 	fflush(stdout);
 
 done:
